/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jws.WebParam;
import javax.jws.WebResult;
import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.Characters;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.ws.Holder;
import javax.xml.ws.WebFault;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.handler.MessageContext;

import org.apache.schemas.yoko.bindings.corba.Alias;
import org.apache.schemas.yoko.bindings.corba.Anonarray;
import org.apache.schemas.yoko.bindings.corba.Anonsequence;
import org.apache.schemas.yoko.bindings.corba.Array;
import org.apache.schemas.yoko.bindings.corba.CaseType;
import org.apache.schemas.yoko.bindings.corba.Enum;
import org.apache.schemas.yoko.bindings.corba.Exception;
import org.apache.schemas.yoko.bindings.corba.MemberType;
import org.apache.schemas.yoko.bindings.corba.ModeType;
import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.ParamType;
import org.apache.schemas.yoko.bindings.corba.RaisesType;
import org.apache.schemas.yoko.bindings.corba.Sequence;
import org.apache.schemas.yoko.bindings.corba.Struct;
import org.apache.schemas.yoko.bindings.corba.Union;
import org.apache.schemas.yoko.bindings.corba.Unionbranch;
import org.apache.yoko.bindings.corba.types.CorbaAnyHandler;
import org.apache.yoko.bindings.corba.types.CorbaArrayHandler;
import org.apache.yoko.bindings.corba.types.CorbaEnumHandler;
import org.apache.yoko.bindings.corba.types.CorbaExceptionHandler;
import org.apache.yoko.bindings.corba.types.CorbaFixedHandler;
import org.apache.yoko.bindings.corba.types.CorbaHandlerUtils;
import org.apache.yoko.bindings.corba.types.CorbaObjectHandler;
import org.apache.yoko.bindings.corba.types.CorbaPrimitiveHandler;
import org.apache.yoko.bindings.corba.types.CorbaSequenceHandler;
import org.apache.yoko.bindings.corba.types.CorbaStructHandler;
import org.apache.yoko.bindings.corba.types.CorbaUnionHandler;
import org.apache.yoko.wsdl.CorbaConstants;
import org.apache.yoko.wsdl.CorbaTypeImpl;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.AbstractBindingImpl;
import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bindings.DataReader;
import org.objectweb.celtix.bindings.DataWriter;
import org.objectweb.celtix.bus.handlers.HandlerChainInvoker;
import org.objectweb.celtix.common.logging.LogUtils;
import org.objectweb.celtix.context.InputStreamMessageContext;
import org.objectweb.celtix.context.ObjectMessageContext;
import org.objectweb.celtix.context.OutputStreamMessageContext;
import org.objectweb.celtix.handlers.HandlerInvoker;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;
import org.objectweb.celtix.wsdl.EndpointReferenceUtils;

import org.omg.CORBA.ORB;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;

public class CorbaBindingImpl extends AbstractBindingImpl {

    private static final Logger LOG = LogUtils.getL7dLogger(CorbaBindingImpl.class);

    private static final String XSI_NAMESPACE_URI = "http://www.w3.org/2001/XMLSchema-instance";
    private static final String XSI_PREFIX = "xsi";
    private static final String XMLSCHEMA_NAMESPACE_URI = "http://www.w3.org/2001/XMLSchema";
    private static final String XMLSCHEMA_PREFIX = "xs";
    
    protected final boolean isServer;
    private Bus bus;
    private EndpointReferenceType endpointRef;
    private ORB orb;
    private List<CorbaTypeMap> typeMaps;

    public CorbaBindingImpl(boolean server) {
        isServer = server;
    }

    public CorbaBindingImpl(Bus b, EndpointReferenceType endpointRefType, ORB theOrb, boolean server) {
        this(server);
        this.bus = b;
        this.endpointRef = endpointRefType;
        this.orb = theOrb;
        
        // initialize the type maps for the binding
        try {
            typeMaps = new ArrayList<CorbaTypeMap>();
            Definition wsdlDef = EndpointReferenceUtils.getWSDLDefinition(bus.getWSDLManager(), 
                                                                          endpointRef);
        
            // first get any typemaps from the main wsdl file
            CorbaUtils.getTypeMapFromDefinition(typeMaps, wsdlDef);
            
            // now check for type maps in any imports
            Map wsdlImports = wsdlDef.getImports();
            Collection importValues = wsdlImports.values();
            for (Iterator iter = importValues.iterator(); iter.hasNext();) {
                List impList = (List)iter.next();
                for (Iterator it = impList.iterator(); it.hasNext();) {
                    Import imp = (Import)it.next();
                    Definition impDef = imp.getDefinition();
                    CorbaUtils.getTypeMapFromDefinition(typeMaps, impDef);
                }
            }
        } catch (java.lang.Exception ex) {
            // Proceed without the type maps.
            LOG.log(Level.INFO, 
                    "Received exception [" + ex + "] while reading typemaps.  Proceeding anyway.");
        }
    }
    
    public Bus getBus() {
        return this.bus;
    }

    public EndpointReferenceType getEndpointReference() {
        return this.endpointRef;
    }

    public List<CorbaTypeMap> getCorbaTypeMaps() {
        return typeMaps;
    }

    // --- AbstractBindingImpl interface ---
    public MessageContext createBindingMessageContext(MessageContext srcCtx) {
        return new CorbaMessageContextImpl(srcCtx);
    }

    public HandlerInvoker createHandlerInvoker() {
        return new HandlerChainInvoker(getHandlerChain(true));
    }

    public void marshal(ObjectMessageContext objContext, MessageContext mc, DataBindingCallback callback) {
        CorbaMessageContext corbaCtx = (CorbaMessageContext) mc;
        corbaCtx.setMessage(new CorbaMessage());

        try {
            DataWriter<XMLEventWriter> writer = callback.createWriter(XMLEventWriter.class);
            if (writer == null) {
                throw new CorbaBindingException("Unable to create data writer");
            }
            
            XMLInputFactory inputFactory = XMLInputFactory.newInstance();
            XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
            outputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);

            // Get information about the operation being invoked from the WSDL definition.
            // We need this to marshal data correctly
            OperationType opElement = 
                CorbaUtils.getCorbaOperationType(callback.getOperationName(), bus, endpointRef);
            if (opElement == null) {
                throw new CorbaBindingException("Unable to find operation definition");
            }

            // Handle the parameters that are given for the operation
            List<ParamType> paramTypes = opElement.getParam();
            boolean hasOutParamReturn = 
                addMarshalParams(callback, paramTypes, corbaCtx, objContext, 
                                 inputFactory, outputFactory, writer);

            // Check to see if we have an input or output message and handle the return type
            // appropriately
            boolean isInputMessage = (Boolean) mc.get(ObjectMessageContext.MESSAGE_INPUT);
            if (!isInputMessage && callback.getWebResult() != null && !hasOutParamReturn) {
                ByteArrayOutputStream outStream = new ByteArrayOutputStream();
                XMLEventWriter evtWriter = outputFactory.createXMLEventWriter(outStream);
                writer.write(objContext.getReturn(), callback.getWebResultQName(), evtWriter);

                ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());
                XMLEventReader evtReader = inputFactory.createXMLEventReader(inStream);
                CorbaObjectHandler obj = readObjectFromStax(evtReader, 
                                                            opElement.getReturn().getIdltype(), 
                                                            false);
                CorbaStreamable streamable = new CorbaStreamable(obj, callback.getWebResultQName());
                corbaCtx.getMessage().setStreamableReturn(streamable);
            } else if (isInputMessage && callback.getWebResult() != null && !hasOutParamReturn) {
                QName retName = callback.getWebResultQName();
                QName retIdlType = opElement.getReturn().getIdltype();
                CorbaObjectHandler obj = 
                    CorbaHandlerUtils.initializeObjectHandler(orb, retName, retIdlType, typeMaps);
                CorbaStreamable streamable = new CorbaStreamable(obj, retName);
                corbaCtx.getMessage().setStreamableReturn(streamable);
            }
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "CORBA marshal exception", ex);
            throw new CorbaBindingException("CORBA Marshal Exception", ex);
        }
    }

    private boolean addMarshalParams(DataBindingCallback callback,
                                     List<ParamType> params,
                                     CorbaMessageContext corbaCtx,
                                     ObjectMessageContext objContext,
                                     XMLInputFactory inputFactory,
                                     XMLOutputFactory outputFactory,
                                     DataWriter<XMLEventWriter> writer) throws java.lang.Exception {
        
        Object[] args = objContext.getMessageObjects();
        Object ret = objContext.getReturn();
        boolean hasOutParamReturn = false;
        
        // We need to check to see if Celtix is using the return type of the operation as the method for
        // passing back the value of a CORBA out parameter.
        if (args != null) {
            hasOutParamReturn = params.size() != args.length;
        } else {
            hasOutParamReturn = params.size() > 0;
        }

        int cbParamCount = 0;
        for (Iterator<ParamType> iter = params.iterator(); iter.hasNext();) {
            ParamType param = iter.next();
            QName elName = null;
            Object partValue = null;
            if (param.getMode().equals(ModeType.OUT) && hasOutParamReturn) {
                // This is actually an OUT parameter to a CORBA application
                WebResult wr = callback.getWebResult();
                elName = new QName(wr.targetNamespace(), wr.name());
                partValue = ret;
            } else {
                WebParam wp = callback.getWebParam(cbParamCount);
                elName = new QName(wp.targetNamespace(), wp.name());
                if (wp.mode() != WebParam.Mode.IN) {
                    partValue = ((Holder)args[cbParamCount]).value;
                } else {
                    partValue = args[cbParamCount];
                }
                cbParamCount++;
            }
            
            ByteArrayOutputStream outStream = new ByteArrayOutputStream();
            XMLEventWriter evtWriter = outputFactory.createXMLEventWriter(outStream);
            writer.write(partValue, elName, evtWriter);

            ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());
            XMLEventReader evtReader = inputFactory.createXMLEventReader(inStream);
            QName idlType = param.getIdltype();
            CorbaObjectHandler obj = null;
            boolean skipRead = false;
            if (isServer) {
                if (param.getMode().equals(ModeType.IN)) {
                    skipRead = true;
                }
            } else if (param.getMode().equals(ModeType.OUT)) {
                skipRead = true;
            } 
            if (skipRead) {
                obj = CorbaHandlerUtils.initializeObjectHandler(orb, elName, idlType, typeMaps);
            } else {
                obj = readObjectFromStax(evtReader, idlType, false);
            }
            CorbaStreamable streamable = new CorbaStreamable(obj, elName);
            ModeType paramMode = param.getMode();
            if (paramMode.value().equals("in")) {
                streamable.setMode(org.omg.CORBA.ARG_IN.value);
            } else if (paramMode.value().equals("inout")) {
                streamable.setMode(org.omg.CORBA.ARG_INOUT.value);
            } else if (paramMode.value().equals("out")) {
                streamable.setMode(org.omg.CORBA.ARG_OUT.value);
            }
            corbaCtx.getMessage().addStreamableArgument(streamable);
        }
        return hasOutParamReturn;
    }

    public void marshalFault(ObjectMessageContext objContext, 
                             MessageContext mc, 
                             DataBindingCallback callback) {
        CorbaMessageContext corbaCtx = (CorbaMessageContext) mc;
        corbaCtx.setMessage(new CorbaMessage());

        // TODO: This method right now assumes that the exception we get is actually an exception that 
        // will be defined in IDL.  What do we want to do if the exception is instead a system exception 
        // that don't have a definition for.  UPDATED: Or will those types of exceptions not get through
        // to this method.  This should be a condition when we insert the exception into the object message 
        // context object that is passed to this method.
        try {
            DataWriter<XMLEventWriter> writer = callback.createWriter(XMLEventWriter.class);
            if (writer == null) {
                throw new WebServiceException("Unable to create data writer");
            }

            // Get information about the operation being invoked from the WSDL definition.
            // We need this to marshal data correctly
            OperationType opElement = 
                CorbaUtils.getCorbaOperationType(callback.getOperationName(), bus, endpointRef);
            if (opElement == null) {
                throw new CorbaBindingException("Unable to find operation definition");
            }

            XMLInputFactory inputFactory = XMLInputFactory.newInstance();
            XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
            outputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
            
            Throwable except = objContext.getException();
            Method faultInfoMethod = except.getClass().getMethod("getFaultInfo");
            Object faultInfo = faultInfoMethod.invoke(except);
            WebFault wfAnnotation = except.getClass().getAnnotation(WebFault.class);
            QName elName = new QName(wfAnnotation.targetNamespace(), except.getClass().getSimpleName());
            
            ByteArrayOutputStream outStream = new ByteArrayOutputStream();
            XMLEventWriter evtWriter = outputFactory.createXMLEventWriter(outStream);
            writer.write(faultInfo, elName, evtWriter);
            
            QName exIdlType = null;
            List<RaisesType> exList = opElement.getRaises();
            for (Iterator<RaisesType> iter = exList.iterator(); iter.hasNext();) {
                // REVISIT: Note that this assumes that exception names need to be unique.  We should make
                // sure that this is really the case.
                RaisesType raises = iter.next();
                if (raises.getException().getLocalPart().equals(elName.getLocalPart())) {
                    exIdlType = raises.getException();
                }
            }
            if (exIdlType == null) {
                LOG.log(Level.INFO, 
                        "Could not find the exception in the raises list.  Must be a system exception.");
                // This could be a system exception?
            } else {
                LOG.log(Level.INFO, "Found exception in the raises list.  Marshalling.");
                ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());
                XMLEventReader evtReader = inputFactory.createXMLEventReader(inStream);
                CorbaObjectHandler obj = readObjectFromStax(evtReader, exIdlType, false);
                CorbaStreamable streamable = new CorbaStreamable(obj, elName);
                corbaCtx.getMessage().setStreamableException(streamable);
            }
            
            
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "CORBA marshalFault exception", ex);
            // TODO: Throw a exception proper exception once implemented
            throw new CorbaBindingException("Method unmarshalFault not implemented");
        }
    }

    public void unmarshal(MessageContext mc, ObjectMessageContext objContext, DataBindingCallback callback) {
        CorbaMessageContext corbaCtx = (CorbaMessageContext) mc;

        try {
            DataReader<XMLEventReader> reader = callback.createReader(XMLEventReader.class);
            if (reader == null) {
                throw new WebServiceException("Unable to create data reader");
            }

            OperationType opElement = 
                CorbaUtils.getCorbaOperationType(callback.getOperationName(), bus, endpointRef);
            if (opElement == null) {
                throw new CorbaBindingException("Unable to find operation definition");
            }

            // We need the list of CORBA parameters to see if we have a single OUT parameter with no
            // return type.  Celtix handles this case a bit differently than everything else
            List<ParamType> params = opElement.getParam();

            // Handle the parameters that are given for the operation
            XMLInputFactory inputFactory = XMLInputFactory.newInstance();
            XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
            outputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
            
            boolean hasOutParamReturn = 
                addUnmarshalParams(callback, params, corbaCtx, objContext, 
                                   inputFactory, outputFactory, reader);
            
            // Check to see if we have an input or output message and handle the return type
            // appropriately
            boolean isInputMessage = (Boolean) mc.get(ObjectMessageContext.MESSAGE_INPUT);
            if (!isInputMessage && callback.getWebResult() != null && !hasOutParamReturn) {
                ByteArrayOutputStream outStream = new ByteArrayOutputStream();
                XMLEventWriter evtWriter = outputFactory.createXMLEventWriter(outStream);
                writeObjectToStax(corbaCtx.getMessage().getStreamableReturn().getObject(), 
                                  evtWriter, XMLEventFactory.newInstance(), false);
                evtWriter.flush();

                ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());
                XMLEventReader evtReader = inputFactory.createXMLEventReader(inStream);
                Object returnValue = reader.read(callback.getWebResultQName(), -1, evtReader);

                objContext.setReturn(returnValue);
            }
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "CORBA unmarshal exception", ex);
            throw new CorbaBindingException("CORBA Unmarshal Exception", ex);
        }
    }
    
    public boolean addUnmarshalParams(DataBindingCallback callback,
                                      List<ParamType> params,
                                      CorbaMessageContext corbaContext,
                                      ObjectMessageContext objContext,
                                      XMLInputFactory inputFactory,
                                      XMLOutputFactory outputFactory, 
                                      DataReader<XMLEventReader> reader) throws java.lang.Exception {
        int cbParamIndex = 0;
        int paramOffset = 0;

        Object[] args = objContext.getMessageObjects();
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        boolean hasOutParamReturn = false;

        // We need to check to see if Celtix is using the return type of the operation as the method for
        // passing back the value of a CORBA out parameter.
        if (args != null) {
            hasOutParamReturn = params.size() != args.length;
        } else {
            hasOutParamReturn = params.size() > 0;
        }

        for (Iterator<ParamType> iter = params.iterator(); iter.hasNext();) {
            ParamType param = iter.next();
            int paramIndex = cbParamIndex + paramOffset;
            
            boolean skipRead = false;
            if (isServer) {
                if (param.getMode().equals(ModeType.OUT)) {
                    skipRead = true;
                }
            } else if (param.getMode().equals(ModeType.IN)) {
                skipRead = true;
            }
            
            Object obj = null;
            if (!skipRead) {
                QName elName = null;
                ByteArrayOutputStream outStream = new ByteArrayOutputStream();
                XMLEventWriter evtWriter = outputFactory.createXMLEventWriter(outStream);
        
                writeObjectToStax(streamables[paramIndex].getObject(), evtWriter, 
                                  XMLEventFactory.newInstance(), false);
                evtWriter.flush();
                
                ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());
                XMLEventReader evtReader = inputFactory.createXMLEventReader(inStream);
                if (param.getMode().equals(ModeType.OUT) && hasOutParamReturn) {
                    WebResult wr = callback.getWebResult();
                    elName = new QName(wr.targetNamespace(), wr.name());
                    obj = reader.read(elName, -1, evtReader);
                    objContext.setReturn(obj);
                    paramOffset = 1;
                } else {
                    WebParam wp = callback.getWebParam(cbParamIndex);
                    elName = new QName(wp.targetNamespace(), wp.name());
                    obj = reader.read(elName, cbParamIndex, evtReader);
                    if (param.getMode().equals(ModeType.INOUT) || param.getMode().equals(ModeType.OUT)) {
                        args[cbParamIndex].getClass().getField("value").set(args[cbParamIndex], obj);
                    } else {
                        args[cbParamIndex] = obj;
                    }
                    cbParamIndex++;
                } 
            } else  {
                if (param.getMode().equals(ModeType.OUT) && hasOutParamReturn) {
                    objContext.setReturn(obj);
                } else if (param.getMode().equals(ModeType.INOUT) || param.getMode().equals(ModeType.OUT)) {
                    args[cbParamIndex].getClass().getField("value").set(args[cbParamIndex], obj);
                } else {
                    args[cbParamIndex] = obj;
                }
                cbParamIndex++;
            }
        }
        return hasOutParamReturn;
    }

    public void unmarshalFault(MessageContext mc, ObjectMessageContext objContext,
                               DataBindingCallback callback) {
        CorbaMessageContext corbaCtx = (CorbaMessageContext) mc;

        try {
            Method currentMethod = (Method)objContext.get(ObjectMessageContext.METHOD_OBJ);
            Class<?>[] methodExs = currentMethod.getExceptionTypes();

            CorbaMessage message = corbaCtx.getMessage();
            CorbaStreamable streamableEx = message.getStreamableException();
            CorbaObjectHandler exObject = streamableEx.getObject();
            String name = exObject.getName().getLocalPart();
            
            Class exWrapperCls = null;
            for (int i = 0; i < methodExs.length; ++i) {
                // Right now, this assumes that all exceptions will have unique simple names.  REVISIT.
                if (name.equals(methodExs[i].getSimpleName())) {
                    exWrapperCls = methodExs[i];
                    break;
                }
            }

            if (exWrapperCls == null) {
                throw new CorbaBindingException("Unable to locate exception type in " 
                                                + "declared exception list");
            }

            try {
                Method faultInfoMethod = exWrapperCls.getMethod("getFaultInfo");
                Class<?> exCls = faultInfoMethod.getReturnType();
                Object yokoEx = exCls.newInstance();

                CorbaExceptionHelper.unmarshalCorbaException((CorbaExceptionHandler)exObject, yokoEx);
                
                Constructor exWrapperCtr = exWrapperCls.getConstructor(String.class, exCls);
                Object exWrapper = exWrapperCtr.newInstance("", yokoEx);

                // the wrapper should always be an instance of throwable
                objContext.setException((Throwable)exWrapper);    

                
            } catch (java.lang.Exception ex) {
                // rethrow to be caught below
                throw ex;
            }

        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "CORBA unmarshalFault exception", ex);
            throw new CorbaBindingException("CORBA unmarshalFault exception", ex);
        }
    }

    public void write(MessageContext msgContext, OutputStreamMessageContext outContext) throws IOException {
        // Complete. This will not be needed for this binding as we are not using Celtix transports
        // which require StreamMessageContexts.
    }

    public void read(InputStreamMessageContext inContext, MessageContext context) throws IOException {
        // Complete. This will not be needed for this binding as we are not using Celtix transports
        // which require StreamMessageContexts.
    }

    public boolean hasFault(MessageContext msgContext) {
        CorbaMessage msg = ((CorbaMessageContext) msgContext).getMessage();
        assert msg != null;
        return msg.getException() != null;
    }

    public void updateMessageContext(MessageContext msgContext) {
        // TODO - provide implementation.
    }

    public CorbaObjectHandler readObjectFromStax(XMLEventReader reader, QName idlType, boolean isNested)
        throws CorbaBindingException {
        CorbaObjectHandler obj = null;
        // Find the first start element in the reader.  This should be the beginning of our object
        try {
            if (!isNested) {
                while (reader.peek().getEventType() != XMLStreamConstants.START_ELEMENT) {
                    reader.nextEvent();
                }
            }
        } catch (XMLStreamException ex) {
            throw new CorbaBindingException("Unable to locate start of object", ex);
        }
        
        TypeCode tc = CorbaUtils.getTypeCode(orb, idlType, typeMaps);
        if (CorbaUtils.isPrimitiveIdlType(idlType)) {
            LOG.log(Level.INFO, "Reading primitive type from XML reader");
            obj = readPrimitiveFromStax(reader, idlType, tc);
        } else {
            switch(tc.kind().value()) {
            case TCKind._tk_alias:
                LOG.log(Level.INFO, "Reading alias type from XML reader");
                obj = readAliasFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_array:
                LOG.log(Level.INFO, "Reading array type from XML reader");
                obj = readArrayFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_enum:
                LOG.log(Level.INFO, "Reading enumeration type from XML reader");
                obj = readEnumFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_except:
                LOG.log(Level.INFO, "Reading exception type from XML reader");
                obj = readExceptionFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_fixed:
                LOG.log(Level.INFO, "Reading fixed type from XML reader");
                obj = readFixedFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_sequence:
                LOG.log(Level.INFO, "Reading sequence type from XML reader");
                obj = readSequenceFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_string:
            case TCKind._tk_wstring:
                // These will be the case if we have anonymous strings
                LOG.log(Level.INFO, "Reading anonymous string from XML reader");
                obj = readAnonStringsFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_struct:
                LOG.log(Level.INFO, "Reading struct type from XML reader");
                obj = readStructFromStax(reader, idlType, tc);
                break;
            case TCKind._tk_union:
                LOG.log(Level.INFO, "Reading union type from XML reader");
                obj = readUnionFromStax(reader, idlType, tc);
                break;
            default:
                throw new CorbaBindingException("Unsupported complex type");
            }
        }

        return obj;
    }
    
    public void writeObjectToStax(CorbaObjectHandler obj, XMLEventWriter writer, 
                                  XMLEventFactory factory, boolean isNested) {
        try {
            QName objName = null;
            if (isNested) {
                objName = new QName(obj.getName().getLocalPart());
            } else {
                objName = obj.getName();
            }

            StartElement startEl = null;
            TypeCode tc = obj.getTypeCode();            
            if (CorbaUtils.isPrimitiveIdlType(obj.getIdlType())) {
                LOG.log(Level.INFO, "Writing primitive type to XML writer");
                
                if (obj.getTypeCodeKind().value() == TCKind._tk_any) {
                    // Any is a special case.  Create an attribute called xsi:type which identifies the type
                    // of object stored in the Any.  Also create a 'dummy' attribute to get the XML schema 
                    // namespace included in the Stax event.  createNamespace() doesn't work for some reason.
                    // TODO: Revisit this.
                    String schemaType = ((CorbaAnyHandler)obj).getSchemaType();
                    QName attrName = new QName(XSI_NAMESPACE_URI, "type", XSI_PREFIX);
                    Attribute attr = factory.createAttribute(attrName, schemaType); 
                    QName nsName = new QName(XMLSCHEMA_NAMESPACE_URI, "namespace", XMLSCHEMA_PREFIX);
                    Attribute ns = factory.createAttribute(nsName, "");
                    List<Attribute> attributes = new ArrayList<Attribute>();
                    attributes.add(attr);
                    attributes.add(ns);
                    
                    startEl = 
                        factory.createStartElement(objName, attributes.iterator(), null);
                } else {
                    startEl = factory.createStartElement(objName, null, null);
                }

                writer.add(startEl);
                writePrimitiveToStax(obj, writer, factory);
            } else {
                startEl = factory.createStartElement(objName, null, null);
                writer.add(startEl);

                switch(tc.kind().value()) {
                // Aliases should never get to this point so we don't need a special alias writer
                case TCKind._tk_array:
                    LOG.log(Level.INFO, "Writing array type to XML writer");
                    writeArrayToStax(obj, writer, factory);
                    break;
                case TCKind._tk_enum:
                    LOG.log(Level.INFO, "Writing enum type to XML writer");
                    writeEnumToStax(obj, writer, factory);
                    break;
                case TCKind._tk_except:
                    LOG.log(Level.INFO, "Writing exception type to XML writer");
                    writeExceptionToStax(obj, writer, factory);
                    break;
                case TCKind._tk_fixed:
                    LOG.log(Level.INFO, "Writing fixed type to XML writer");
                    writeFixedToStax(obj, writer, factory);
                    break;
                case TCKind._tk_sequence:
                    LOG.log(Level.INFO, "Writing sequence type to XML writer");
                    writeSequenceToStax(obj, writer, factory);
                    break;
                case TCKind._tk_struct:
                    LOG.log(Level.INFO, "Writing struct type to XML writer");
                    writeStructToStax(obj, writer, factory);
                    break;
                case TCKind._tk_union:
                    LOG.log(Level.INFO, "Writing union type to XML writer");
                    writeUnionToStax(obj, writer, factory);
                    break;
                case TCKind._tk_string:
                case TCKind._tk_wstring:
                    LOG.log(Level.INFO, "Wrting anonymous string to XML writer");
                    writePrimitiveToStax(obj, writer, factory);
                    break;
                default:
                    throw new CorbaBindingException("Unsupported complex type");
                }
            }            
            
            EndElement endEl = factory.createEndElement(objName, null);
            writer.add(endEl);
        } catch (XMLStreamException ex) {
            LOG.log(Level.SEVERE, "STAX exception while writing to STAX event writer: " + ex.toString());
            LOG.log(Level.SEVERE, "Object being written: " + obj);
            throw new CorbaBindingException("writeObjectToStax STAX exception", ex);
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Non-STAX exception while writing to STAX event writer: " + ex.toString());
            LOG.log(Level.SEVERE, "Object being written: " + obj);
            throw new CorbaBindingException("writeObjectToStax exception", ex);
        }
    }

    public CorbaObjectHandler readPrimitiveFromStax(XMLEventReader reader, QName idlType, TypeCode tc) 
        throws CorbaBindingException {
        try {
            StartElement startEl = reader.nextEvent().asStartElement();
            Characters charEl = reader.nextEvent().asCharacters();
            reader.nextEvent().asEndElement();
            if (tc.kind().value() == TCKind._tk_any) {
                Iterator iter = startEl.getAttributes();
                String anySchemaType = null;
                while (iter.hasNext()) {
                    Attribute attr = (Attribute)iter.next();
                    QName attrName = attr.getName();
                    if (attrName.getNamespaceURI().equals(XSI_NAMESPACE_URI)
                        && attrName.getLocalPart().equals("type")) {
                        anySchemaType = attr.getValue();
                        break;
                    }
                }
                           
                CorbaAnyHandler obj = new CorbaAnyHandler(startEl.getName(), idlType, tc, null);
                obj.setValueFromData(orb, charEl.getData(), anySchemaType);
                return obj;
                
            } else {
                CorbaPrimitiveHandler obj = new CorbaPrimitiveHandler(startEl.getName(), idlType, tc, null);
                obj.setValueFromData(charEl.getData());
                return obj;
            }
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Error: Object not correctly defined in the XML reader");
            throw new CorbaBindingException("Invalid XML event read", ex);
        }
    }
    
    public void writePrimitiveToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        Characters charEvt = null;
        if (obj.getTypeCodeKind() == TCKind.tk_any) {
            charEvt = factory.createCharacters(((CorbaAnyHandler)obj).getValueData());
        } else {
            charEvt = factory.createCharacters(((CorbaPrimitiveHandler)obj).getValueData());
        }
        writer.add(charEvt);
    }

    public CorbaObjectHandler readAliasFromStax(XMLEventReader reader, QName idlType, TypeCode tc) {
        CorbaObjectHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);

        // We need to create an ObjectHandler from the base type of the typedef.  So we recursively call
        // the readObjectFromStax call.  However, this is not a true nested read so still pass false as the 
        // nested flag.
        try {
            Alias aliasType = (Alias)typeDefinition;
            obj = readObjectFromStax(reader, aliasType.getBasetype(), false);
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading alias corba type", ex);
        }
        return obj;
    }
    
    public CorbaObjectHandler readAnonStringsFromStax(XMLEventReader reader, QName idlType, TypeCode tc) {
        CorbaPrimitiveHandler obj = null;
        
        try {
            StartElement startEl = reader.nextEvent().asStartElement();
            Characters charEl = reader.nextEvent().asCharacters();
            reader.nextEvent().asEndElement();
            
            if (tc.kind().value() == TCKind._tk_string) {
                obj = 
                    new CorbaPrimitiveHandler(startEl.getName(), CorbaConstants.NT_CORBA_STRING, tc, null);
            } else {
                obj = 
                    new CorbaPrimitiveHandler(startEl.getName(), CorbaConstants.NT_CORBA_WSTRING, tc, null);
            }
            obj.setValueFromData(charEl.getData());
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Error: Object not correctly defined in the XML reader");
            throw new CorbaBindingException("Invalid XML event read", ex);
        }
        return obj;
    }
    
    // Anonymous strings do not need an equivalent write - the primitive handler write is sufficuent. 
    
    public CorbaObjectHandler readArrayFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaArrayHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            QName arrayElemType = null;
            long arrayBound = 0;
            // Arrays and anonymous arrays can be handled by the same method
            if (typeDefinition instanceof Anonarray) {
                Anonarray anonArrayType = (Anonarray)typeDefinition;
                arrayElemType = anonArrayType.getElemtype();
                arrayBound = anonArrayType.getBound();
            } else {
                Array arrayType = (Array)typeDefinition;
                arrayElemType = arrayType.getElemtype();
                arrayBound = arrayType.getBound();
            }
            StartElement arrayStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaArrayHandler(arrayStartEl.getName(), idlType, tc, typeDefinition);
            for (int i = 0; i < arrayBound; ++i) {
                CorbaObjectHandler element = readObjectFromStax(reader, arrayElemType, true);
                obj.addElement(element);
            }
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading array corba type", ex);
        }
        
        return obj;
    }

    public void writeArrayToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaArrayHandler arrayHandler = (CorbaArrayHandler)obj;
        List<CorbaObjectHandler> elements = arrayHandler.getElements();
        for (Iterator<CorbaObjectHandler> iter = elements.iterator(); iter.hasNext();) {
            writeObjectToStax(iter.next(), writer, factory, true);
        }
    }
    
    public CorbaObjectHandler readEnumFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaEnumHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            Enum enumType = (Enum)typeDefinition;
            StartElement enumStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaEnumHandler(enumStartEl.getName(), idlType, tc, enumType);
            Characters enumCharEl = reader.nextEvent().asCharacters();
            obj.setValue(enumCharEl.getData());
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading enum corba type", ex);
        }
        
        return obj;
    }
    
    public void writeEnumToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaEnumHandler enumHandler = (CorbaEnumHandler)obj;
        Characters charEvt = factory.createCharacters(enumHandler.getValue());
        writer.add(charEvt);
    }

    public CorbaObjectHandler readExceptionFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaExceptionHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            Exception exType = (Exception)typeDefinition;
            StartElement exStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaExceptionHandler(exStartEl.getName(), idlType, tc, exType);
            List<MemberType> exMembers = exType.getMember();
            for (Iterator<MemberType> iter = exMembers.iterator(); iter.hasNext();) {
                MemberType m = iter.next();
                CorbaObjectHandler member = readObjectFromStax(reader, m.getIdltype(), true);
                obj.addMember(member);
            }
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading exception corba type", ex);
        }
        
        return obj;
    }

    public void writeExceptionToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaExceptionHandler exObj = (CorbaExceptionHandler)obj;
        List<CorbaObjectHandler> elements = exObj.getMembers();
        for (Iterator<CorbaObjectHandler> iter = elements.iterator(); iter.hasNext();) {
            writeObjectToStax(iter.next(), writer, factory, true);
        }
    }

    public CorbaObjectHandler readSequenceFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaSequenceHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            QName seqElementType = null;
            long bound = 0;
            if (typeDefinition instanceof Anonsequence) {
                Anonsequence anonSeqType = (Anonsequence)typeDefinition;
                seqElementType = anonSeqType.getElemtype();
                bound = anonSeqType.getBound();
            } else {
                Sequence seqType = (Sequence)typeDefinition;
                seqElementType = seqType.getElemtype();
                bound = seqType.getBound();
            }
            StartElement seqStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaSequenceHandler(seqStartEl.getName(), idlType, tc, typeDefinition);
            if (bound == 0) {
                LOG.log(Level.INFO, "Unbounded sequence found");
                while (reader.peek().getEventType() == XMLStreamConstants.START_ELEMENT) {
                    CorbaObjectHandler element = readObjectFromStax(reader, seqElementType, true);
                    obj.addElement(element);
                }
            } else {
                LOG.log(Level.INFO, "Bounded sequence found");
                for (long i = 0; i < bound; ++i) {
                    CorbaObjectHandler element = readObjectFromStax(reader, seqElementType, true);
                    obj.addElement(element);
                }
            }
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading sequence corba type", ex);
        }
        
        return obj;
    }

    public void writeSequenceToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaSequenceHandler seqHandler = (CorbaSequenceHandler)obj;
        List<CorbaObjectHandler> elements = seqHandler.getElements();
        for (Iterator<CorbaObjectHandler> iter = elements.iterator(); iter.hasNext();) {
            writeObjectToStax(iter.next(), writer, factory, true);
        }
    }

    public CorbaObjectHandler readStructFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaStructHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            Struct structType = (Struct)typeDefinition;
            StartElement structStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaStructHandler(structStartEl.getName(), idlType, tc, structType);
            List<MemberType> structMembers = structType.getMember();
            for (Iterator<MemberType> iter = structMembers.iterator(); iter.hasNext();) {
                MemberType m = iter.next();
                CorbaObjectHandler member = readObjectFromStax(reader, m.getIdltype(), true);
                obj.addMember(member);
            }
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading struct corba type", ex);
        }
        
        return obj;
    }

    public void writeStructToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaStructHandler structHandler = (CorbaStructHandler)obj;
        List<CorbaObjectHandler> elements = structHandler.getMembers();
        for (Iterator<CorbaObjectHandler> iter = elements.iterator(); iter.hasNext();) {
            writeObjectToStax(iter.next(), writer, factory, true);
        }
    }


    public CorbaObjectHandler readUnionFromStax(XMLEventReader reader, QName idlType, TypeCode tc)
        throws CorbaBindingException {
        CorbaUnionHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            Union unionType = (Union)typeDefinition;
            StartElement unionStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaUnionHandler(unionStartEl.getName(), idlType, tc, unionType);

            // Read the discriminator value from the reader.  We can't rely on the value from 
            // Stax since it may be a default value if the caller has forgotten to explicitly
            // set the discriminator in their union class.  This could cause all sorts of 
            // problems when we try create the CORBA union.  For now, we'll just read and ignore.
            // TODO: Revisit this if we want to make this assumption.
            reader.nextEvent().asStartElement().getName();
            reader.nextEvent().asCharacters();
            reader.nextEvent().asEndElement();
            
            // Build the entire union with all branches, etc.  Then read info from the XML Event Reader
            StartElement branchElement = reader.peek().asStartElement();
            String branchName = branchElement.getName().getLocalPart();
            List<Unionbranch> branches = unionType.getUnionbranch();
            for (Iterator<Unionbranch> iter = branches.iterator(); iter.hasNext();) {
                Unionbranch branch = iter.next();
                CorbaObjectHandler branchObj = null;
                if (branch.getName().equals(branchName)) {
                    branchObj = readObjectFromStax(reader, branch.getIdltype(), true);
                    // We also need to set the discriminator since this is the branch with the actual
                    // union value
                    CorbaObjectHandler discObj = 
                        CorbaHandlerUtils.createTypeHandler(orb, new QName("discriminator"),
                                                            unionType.getDiscriminator(), typeMaps);
                    obj.setDiscriminator(discObj);

                    // Determine the value of the discriminator.  
                    List<CaseType> branchCases = branch.getCase();
                    String discValue = null;
                    if (branchCases.size() != 0) {
                        // This represents a union case.  Set the discriminator based on the first
                        // label value associated with the branch (since we don't have this information)
                        // from the Stax representation of the Celtix object).
                        CaseType caseLabel = branchCases.get(0);
                        discValue = caseLabel.getLabel();
                    } else {
                        // This represents the default case.  Since Celtix does not provide a 
                        // discriminator value, we'll create one.
                        discValue = obj.createDefaultDiscriminatorLabel();
                    }
                    obj.setDiscriminatorValueFromData(discValue);
                    obj.setValue(branchName, branchObj);
                } else {
                    // Create an object holder with no value
                    branchObj = CorbaHandlerUtils.createTypeHandler(orb, new QName(branch.getName()),
                                                                    branch.getIdltype(), typeMaps);
                }
                obj.addCase(branchObj);
            }
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading union corba type", ex);
        }
        
        return obj;
    }
    
    public void writeUnionToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory)
        throws XMLStreamException {
        CorbaUnionHandler unionHandler = (CorbaUnionHandler)obj;
        CorbaObjectHandler discValue = unionHandler.getDiscriminator();
        writeObjectToStax(discValue, writer, factory, true);
        CorbaObjectHandler unionValue = unionHandler.getValue();
        writeObjectToStax(unionValue, writer, factory, true);
    }
    
    public CorbaObjectHandler readFixedFromStax(XMLEventReader reader, QName idlType, TypeCode tc) 
        throws CorbaBindingException {
        CorbaFixedHandler obj = null;
        CorbaTypeImpl typeDefinition = CorbaUtils.getCorbaType(idlType, typeMaps);
        
        try {
            StartElement fixedStartEl = reader.nextEvent().asStartElement();
            obj = new CorbaFixedHandler(fixedStartEl.getName(), idlType, tc, typeDefinition);
            Characters fixedCharEl = reader.nextEvent().asCharacters();
            obj.setValueFromData(fixedCharEl.getData());
            reader.nextEvent().asEndElement();
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Received exception while reading object of type " + idlType);
            throw new CorbaBindingException("Error while reading fixed corba type", ex);
        }
        return obj;
    }
    
    public void writeFixedToStax(CorbaObjectHandler obj, XMLEventWriter writer, XMLEventFactory factory) 
        throws XMLStreamException {
        CorbaFixedHandler fixedHandler = (CorbaFixedHandler)obj;
        Characters charEvt = factory.createCharacters(fixedHandler.getValueData());
        writer.add(charEvt);
    }
}
