/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.handler.MessageContext;

import org.apache.schemas.yoko.bindings.corba.BindingType;
import org.apache.schemas.yoko.bindings.corba.ModeType;
import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.ParamType;
import org.apache.yoko.bindings.corba.types.CorbaHandlerUtils;
import org.apache.yoko.bindings.corba.types.CorbaObjectHandler;

import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bindings.ServerBindingEndpointCallback;
import org.objectweb.celtix.bindings.ServerDataBindingCallback;
import org.objectweb.celtix.context.GenericMessageContext;
import org.objectweb.celtix.context.ObjectMessageContext;
import org.objectweb.celtix.context.WebServiceContextImpl;

import org.omg.CORBA.Any;
import org.omg.CORBA.NVList;
import org.omg.CORBA.ORB;
import org.omg.CORBA.ServerRequest;
import org.omg.PortableServer.DynamicImplementation;
import org.omg.PortableServer.POA;

public class CorbaDSIServant extends DynamicImplementation {

    private final ORB orb;
    private final POA servantPOA;
    private final CorbaServerBinding serverBinding;
    private ServerBindingEndpointCallback sbeCallback;
    private List<String> interfaces;

    public CorbaDSIServant(ORB theOrb, POA poa, CorbaServerBinding binding,
                           ServerBindingEndpointCallback epCallback) {
        orb = theOrb;
        servantPOA = poa;
        serverBinding = binding;
        sbeCallback = epCallback;

        // Get the list of interfaces that this servant will support
        try {
            BindingType bindType = CorbaUtils.getCorbaBindingType(binding.getBus(), 
                                                                  binding.getEndpointReference());
            if (bindType == null) {
                throw new CorbaBindingException("Unable to determine binding information");
            }

            List<String> bases = bindType.getBases();
            interfaces = new ArrayList<String>();
            interfaces.add(bindType.getRepositoryID());
            for (Iterator<String> iter = bases.iterator(); iter.hasNext();) {
                interfaces.add(iter.next());
            }
        } catch (java.lang.Exception ex) {
            // TODO: For now, this exception is consumed. Revisit this to see if
            // it is the best way to handle this.
        }
    }

    public void invoke(ServerRequest request) throws CorbaBindingException {
        try {
            CorbaMessageContext corbaCtx = 
                (CorbaMessageContext)serverBinding.getBindingImpl().
                    createBindingMessageContext(new GenericMessageContext());
            corbaCtx.setMessage(new CorbaMessage());
            corbaCtx.put(ObjectMessageContext.MESSAGE_INPUT, Boolean.TRUE);

            QName operationName = new QName(request.operation());
            corbaCtx.put(MessageContext.WSDL_OPERATION, operationName);

            ObjectMessageContext objectCtx = serverBinding.createObjectContext();
            DataBindingCallback.Mode mode = sbeCallback.getServiceMode();
            DataBindingCallback callback = 
                sbeCallback.getDataBindingCallback((QName)corbaCtx.get(MessageContext.WSDL_OPERATION),
                                                   objectCtx, 
                                                   mode);
            callback.initObjectContext(objectCtx);
            objectCtx.putAll(corbaCtx);

            // Get the WSDL definition for the operation being invoked 
            // TODO: Verify that two operations can't be overloaded.
            OperationType opElement = CorbaUtils.getCorbaOperationType(operationName.getLocalPart(), 
                                                                       serverBinding.getBus(),
                                                                       serverBinding.getEndpointReference());
            if (opElement == null) {
                throw new CorbaBindingException("Unable to locate operation type definition");
            }

            List<ParamType> paramTypes = opElement.getParam();
            List<CorbaTypeMap> typeMaps = 
                ((CorbaBindingImpl)serverBinding.getBindingImpl()).getCorbaTypeMaps();

            // Build a list of streamables that correspond to the values that should be contained in the
            // argument list of the ServerRequest, build the correct NVList and retreive the arguments
            CorbaStreamable[] arguments = new CorbaStreamable[paramTypes.size()];
            NVList list = prepareDIIArgsList(corbaCtx, objectCtx, arguments, paramTypes, typeMaps, callback);
            request.arguments(list);

            serverBinding.getBindingImpl().unmarshal(corbaCtx, objectCtx, callback);

            boolean isOneWay = callback.isOneWay();
            doCeltixInvocation(callback, objectCtx);

            corbaCtx.put(ObjectMessageContext.MESSAGE_INPUT, Boolean.FALSE);
            if (!isOneWay) {
                if (objectCtx.getException() != null) {
                    serverBinding.getBindingImpl().marshalFault(objectCtx, corbaCtx, callback);
                    Any exAny = orb.create_any();
                    exAny.insert_Streamable(corbaCtx.getMessage().getStreamableException());
                    request.set_exception(exAny);
                } else {
                    serverBinding.getBindingImpl().marshal(objectCtx, corbaCtx, callback);

                    arguments = corbaCtx.getMessage().getStreamableArguments();
                    for (int i = 0; i < arguments.length; ++i) {
                        if (list.item(i).flags() != org.omg.CORBA.ARG_IN.value) {
                            list.item(i).value().insert_Streamable(arguments[i]);
                        }
                    }

                    CorbaStreamable resultValue = corbaCtx.getMessage().getStreamableReturn();
                    if (resultValue != null) {
                        Any resultAny = orb.create_any();
                        resultAny.insert_Streamable(resultValue);
                        request.set_result(resultAny);
                    }
                }
            }

        } catch (java.lang.Exception ex) {
            throw new CorbaBindingException("Exception during invoke", ex);
        }
    }
    
    public NVList prepareDIIArgsList(CorbaMessageContext corbaCtx, ObjectMessageContext objContext, 
                                     CorbaStreamable[] streamables, List<ParamType> params,  
                                     List<CorbaTypeMap> typeMaps, DataBindingCallback callback) {
        NVList list = orb.create_list(streamables.length);

        // We need to check to see if Celtix is using the return type of the operation as the method for
        // passing back the value of a CORBA out parameter.
        Object[] args = objContext.getMessageObjects();
        boolean hasOutParamReturn = false;
        if (args != null) {
            hasOutParamReturn = params.size() != args.length;
        } else {
            hasOutParamReturn = params.size() > 0;
        }
        
        int paramOffset = 0;
        int cbParamCount = 0;
        for (Iterator<ParamType> iter = params.iterator(); iter.hasNext();) {
            int index = cbParamCount + paramOffset;
            ParamType pType = iter.next();
            QName paramName = null;
            if (pType.getMode().equals(ModeType.OUT) && hasOutParamReturn) {
                paramName = new QName(callback.getWebResult().targetNamespace(),
                                      callback.getWebResult().name());
                paramOffset = 1;
            } else {
                paramName = new QName(callback.getWebParam(cbParamCount).targetNamespace(),
                                      callback.getWebParam(cbParamCount).name());
                cbParamCount++;
            }
            QName paramIdlType = pType.getIdltype();
            ModeType paramMode = pType.getMode();

            CorbaObjectHandler obj = 
                CorbaHandlerUtils.initializeObjectHandler(orb, paramName, paramIdlType, typeMaps);

            streamables[index] = new CorbaStreamable(obj, paramName);
            if (paramMode.value().equals("in")) {
                streamables[index].setMode(org.omg.CORBA.ARG_IN.value);
            } else if (paramMode.value().equals("out")) {
                streamables[index].setMode(org.omg.CORBA.ARG_OUT.value);
            } else {
                streamables[index].setMode(org.omg.CORBA.ARG_INOUT.value);
            }

            Any value = orb.create_any();
            value.insert_Streamable(streamables[index]);
            list.add_value(streamables[index].getName(), value, streamables[index].getMode());
            corbaCtx.getMessage().addStreamableArgument(streamables[index]);
        }
        return list;
    }

    public void doCeltixInvocation(DataBindingCallback callback, ObjectMessageContext objectCtx) {
        try {
            ServerDataBindingCallback method = (ServerDataBindingCallback) callback;
            new WebServiceContextImpl(objectCtx);
            method.invoke(objectCtx);
        } catch (WebServiceException ex) {
            Throwable cause = ex.getCause();
            if (cause != null) {
                objectCtx.setException(cause);
            } else {
                objectCtx.setException(ex);
            }
        } catch (InvocationTargetException ex) {
            Throwable cause = ex.getCause();
            if (cause != null) {
                objectCtx.setException(cause);
            } else {
                objectCtx.setException(ex);
            }
        }
    }
    
    public String[] _all_interfaces(POA poa, byte[] objectId) {
        return interfaces.toArray(new String[interfaces.size()]);
    }

    public POA _default_POA() {
        return servantPOA;
    }
}
