/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba.extensions;

import javax.wsdl.Definition;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.ExtensionDeserializer;
import javax.wsdl.extensions.ExtensionRegistry;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.apache.schemas.yoko.bindings.corba.Alias;
import org.apache.schemas.yoko.bindings.corba.Anonarray;
import org.apache.schemas.yoko.bindings.corba.Anonfixed;
import org.apache.schemas.yoko.bindings.corba.Anonsequence;
import org.apache.schemas.yoko.bindings.corba.Anonstring;
import org.apache.schemas.yoko.bindings.corba.Array;
import org.apache.schemas.yoko.bindings.corba.CaseType;
import org.apache.schemas.yoko.bindings.corba.Const;
import org.apache.schemas.yoko.bindings.corba.Enum;
import org.apache.schemas.yoko.bindings.corba.Enumerator;
import org.apache.schemas.yoko.bindings.corba.Exception;
import org.apache.schemas.yoko.bindings.corba.Fixed;
import org.apache.schemas.yoko.bindings.corba.MemberType;
import org.apache.schemas.yoko.bindings.corba.Sequence;
import org.apache.schemas.yoko.bindings.corba.Struct;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;
import org.apache.schemas.yoko.bindings.corba.Union;
import org.apache.schemas.yoko.bindings.corba.Unionbranch;
import org.apache.yoko.wsdl.CorbaConstants;

public class CorbaTypesExtensionHelper implements ExtensionDeserializer {

    public CorbaTypesExtensionHelper() {
    }

    public static void addExtension(ExtensionRegistry registry, 
                                    Class parentType,
                                    String element) {
        QName extQName = new QName(CorbaConstants.NU_WSDL_CORBA, element, CorbaConstants.NP_WSDL_CORBA);
        CorbaTypesExtensionHelper helper = new CorbaTypesExtensionHelper();
        
        registry.registerDeserializer(parentType, extQName, helper);
    }

    public ExtensibilityElement unmarshall(Class parentType, QName elementType,
                                           Element el, Definition def,
                                           ExtensionRegistry extReg)
        throws WSDLException {
        TypeMappingType mappingType = new TypeMappingType();

        // Store the target namespace for the types
        NamedNodeMap mappingAttributes = el.getAttributes();
        for (int i = 0; i < mappingAttributes.getLength(); ++i) {
            if (mappingAttributes.item(i).getNodeName().equals("targetNamespace")) {
                mappingType.setTargetNamespace(mappingAttributes.item(i).getNodeValue());
            }
        }

        // Get the types that are stored in this typemapping
        NodeList typeChildNodes = el.getChildNodes();
        for (int i = 0; i < typeChildNodes.getLength(); ++i) {
            Node currentNode = typeChildNodes.item(i);

            if (currentNode.getNodeName().equals("corba:const")) {
                Const constType = getConstDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(constType);
            } else if (currentNode.getNodeName().equals("corba:enum")) {
                Enum enumType = getEnumDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(enumType);
            } else if (currentNode.getNodeName().equals("corba:struct")) {
                Struct structType = getStructDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(structType);
            } else if (currentNode.getNodeName().equals("corba:exception")) {
                Exception exceptType = getExceptionDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(exceptType);
            } else if (currentNode.getNodeName().equals("corba:fixed")) {
                Fixed fixedType = getFixedDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(fixedType);
            } else if (currentNode.getNodeName().equals("corba:union")) {
                Union unionType = getUnionDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(unionType);
            } else if (currentNode.getNodeName().equals("corba:alias")) {
                Alias aliasType = getAliasDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(aliasType);
            } else if (currentNode.getNodeName().equals("corba:array")) {
                Array arrayType = getArrayDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(arrayType);
            } else if (currentNode.getNodeName().equals("corba:sequence")) {
                Sequence sequenceType = getSequenceDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(sequenceType);
            } else if (currentNode.getNodeName().equals("corba:anonstring")) {
                Anonstring anonstringType = getAnonStringDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(anonstringType);
            } else if (currentNode.getNodeName().equals("corba:anonwstring")) {
                Anonstring anonstringType = getAnonWStringDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(anonstringType);
            } else if (currentNode.getNodeName().equals("corba:anonfixed")) {
                Anonfixed anonfixedType = getAnonFixedDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(anonfixedType);
            } else if (currentNode.getNodeName().equals("corba:anonsequence")) {
                Anonsequence anonsequenceType = getAnonSequenceDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(anonsequenceType);
            } else if (currentNode.getNodeName().equals("corba:anonarray")) {
                Anonarray anonarrayType = getAnonArrayDefinition(currentNode, def);
                mappingType.getStructOrExceptionOrUnion().add(anonarrayType);
            } else {
                // TODO: How do we want to handle this?
            }
        }
        return mappingType;
    }

    public Const getConstDefinition(Node node, Definition def) {
        Const constType = new Const();
        NamedNodeMap constAttributes = node.getAttributes();

        // Store information about the const
        for (int i = 0; i < constAttributes.getLength(); ++i) {
            if (constAttributes.item(i).getNodeName().equals("name")) {
                constType.setName(constAttributes.item(i).getNodeValue());
            } else if (constAttributes.item(i).getNodeName().equals("value")) {
                constType.setValue(constAttributes.item(i).getNodeValue());
            } else if (constAttributes.item(i).getNodeName().equals("idltype")) {
                String idlType = constAttributes.item(i).getNodeValue();
                int seperatorIndex = idlType.indexOf(':');
                String prefix = idlType.substring(0, seperatorIndex);
                String localPart = idlType.substring(seperatorIndex + 1, idlType.length());
                constType.setIdltype(new QName(def.getNamespace(prefix), localPart, prefix));
            }
        }
        return constType;
    }

    public Enum getEnumDefinition(Node node, Definition def) {
        Enum enumType = new Enum();
        NamedNodeMap enumAttributes = node.getAttributes();
        
        // Store information about the enum
        for (int i = 0; i < enumAttributes.getLength(); ++i) {
            if (enumAttributes.item(i).getNodeName().equals("name")) {
                enumType.setName(enumAttributes.item(i).getNodeValue());
            } else if (enumAttributes.item(i).getNodeName().equals("repositoryID")) {
                enumType.setRepositoryID(enumAttributes.item(i).getNodeValue());
            }
        }

        // Store information about the enumerations
        NodeList enumChildNodes = node.getChildNodes();
        for (int i = 0; i < enumChildNodes.getLength(); ++i) {
            Node currentNode = enumChildNodes.item(i);
            
            // The spec uses corba:enumeration but the tool we have uses corba:enumerator.
            // REVISIT: Which one is correct?
            //if (currentNode.getNodeName().equals("corba:enumeration")) {
            if (currentNode.getNodeName().equals("corba:enumerator")) {
                Enumerator enumerator = new Enumerator();
                NamedNodeMap enumeratorAttrs = currentNode.getAttributes();

                for (int j = 0; j < enumeratorAttrs.getLength(); ++j) {
                    Node enumeratorAttrNode = enumeratorAttrs.item(j);
                    if (enumeratorAttrNode.getNodeName().equals("value")) {
                        enumerator.setValue(enumeratorAttrNode.getNodeValue());
                    }
                }
                enumType.getEnumerator().add(enumerator);
            }
        }

        return enumType;
    }

    public Struct getStructDefinition(Node node, Definition def) {
        Struct structType = new Struct();
        NamedNodeMap structAttributes = node.getAttributes();

        // Store information about the struct
        for (int i = 0; i < structAttributes.getLength(); ++i) {
            if (structAttributes.item(i).getNodeName().equals("name")) {
                structType.setName(structAttributes.item(i).getNodeValue());
            } else if (structAttributes.item(i).getNodeName().equals("repositoryID")) {
                structType.setRepositoryID(structAttributes.item(i).getNodeValue());
            }
        }

        // Store information about the structs members
        NodeList structChildNodes = node.getChildNodes();
        for (int i = 0; i < structChildNodes.getLength(); ++i) {
            Node currentNode = structChildNodes.item(i);

            if (currentNode.getNodeName().equals("corba:member")) {
                MemberType member = new MemberType();
                NamedNodeMap memberAttributes = currentNode.getAttributes();

                for (int j = 0; j < memberAttributes.getLength(); ++j) {
                    Node memberAttrNode = memberAttributes.item(j);
                    if (memberAttrNode.getNodeName().equals("name")) {
                        member.setName(memberAttrNode.getNodeValue());
                    } else if (memberAttrNode.getNodeName().equals("idltype")) {
                        String idlType = memberAttrNode.getNodeValue();
                        int seperatorIndex = idlType.indexOf(':');
                        String prefix = idlType.substring(0, seperatorIndex);
                        String localPart = idlType.substring(seperatorIndex + 1, idlType.length());
                        member.setIdltype(new QName(def.getNamespace(prefix), localPart, prefix));
                    }
                }
                structType.getMember().add(member);
            }
        }
        return structType;
    }

    public Exception getExceptionDefinition(Node node, Definition def) {
        Exception exceptType = new Exception();
        NamedNodeMap exceptAttributes = node.getAttributes();

        // Store information about the exception
        for (int i = 0; i < exceptAttributes.getLength(); ++i) {
            if (exceptAttributes.item(i).getNodeName().equals("name")) {
                exceptType.setName(exceptAttributes.item(i).getNodeValue());
            } else if (exceptAttributes.item(i).getNodeName().equals("repositoryID")) {
                exceptType.setRepositoryID(exceptAttributes.item(i).getNodeValue());
            }
        }

        // Store information about the exceptions members
        NodeList exceptChildNodes = node.getChildNodes();
        for (int i = 0; i < exceptChildNodes.getLength(); ++i) {
            Node currentNode = exceptChildNodes.item(i);

            if (currentNode.getNodeName().equals("corba:member")) {
                MemberType member = new MemberType();
                NamedNodeMap memberAttributes = currentNode.getAttributes();

                for (int j = 0; j < memberAttributes.getLength(); ++j) {
                    Node memberAttrNode = memberAttributes.item(j);
                    if (memberAttrNode.getNodeName().equals("name")) {
                        member.setName(memberAttrNode.getNodeValue());
                    } else if (memberAttrNode.getNodeName().equals("idltype")) {
                        String idlType = memberAttrNode.getNodeValue();
                        int seperatorIndex = idlType.indexOf(':');
                        String prefix = idlType.substring(0, seperatorIndex);
                        String localPart = idlType.substring(seperatorIndex + 1, idlType.length());
                        member.setIdltype(new QName(def.getNamespace(prefix), localPart, prefix));
                    }
                }
                exceptType.getMember().add(member);
            }
        }
        return exceptType;
    }

    public Fixed getFixedDefinition(Node node, Definition def) {
        Fixed fixedType = new Fixed();

        // Store information about the fixed type
        NamedNodeMap fixedAttributes = node.getAttributes();
        for (int i = 0; i < fixedAttributes.getLength(); ++i) {
            if (fixedAttributes.item(i).getNodeName().equals("name")) {
                fixedType.setName(fixedAttributes.item(i).getNodeValue());
            } else if (fixedAttributes.item(i).getNodeName().equals("repositoryID")) {
                fixedType.setRepositoryID(fixedAttributes.item(i).getNodeValue());
            } else if (fixedAttributes.item(i).getNodeName().equals("digits")) {
                fixedType.setDigits(new Long(fixedAttributes.item(i).getNodeValue()));
            } else if (fixedAttributes.item(i).getNodeName().equals("scale")) {
                fixedType.setScale(new Long(fixedAttributes.item(i).getNodeValue()));
            }
        }
        return fixedType;
    }

    public Union getUnionDefinition(Node node, Definition def) {
        Union unionType = new Union();
        NamedNodeMap unionAttributes = node.getAttributes();

        for (int i = 0; i < unionAttributes.getLength(); ++i) {
            if (unionAttributes.item(i).getNodeName().equals("name")) {
                unionType.setName(unionAttributes.item(i).getNodeValue());
            } else if (unionAttributes.item(i).getNodeName().equals("repositoryID")) {
                unionType.setRepositoryID(unionAttributes.item(i).getNodeValue());
            } else if (unionAttributes.item(i).getNodeName().equals("discriminator")) {
                String idlType = unionAttributes.item(i).getNodeValue();
                int seperatorIndex = idlType.indexOf(':');
                String prefix = idlType.substring(0, seperatorIndex);
                String localPart = idlType.substring(seperatorIndex + 1, idlType.length());
                unionType.setDiscriminator(new QName(def.getNamespace(prefix), localPart, prefix));
            }
        }

        NodeList unionChildNodes = node.getChildNodes();
        for (int i = 0; i < unionChildNodes.getLength(); ++i) {
            Node currentNode = unionChildNodes.item(i);
            
            if (currentNode.getNodeName().equals("corba:unionbranch")) {
                Unionbranch branch = getUnionBranch(currentNode, def);
                unionType.getUnionbranch().add(branch);
            }
        }
        
        return unionType;
    }
    
    public Unionbranch getUnionBranch(Node node, Definition def) {
        Unionbranch branchType = new Unionbranch();
        NamedNodeMap branchAttributes = node.getAttributes();
        
        for (int i = 0; i < branchAttributes.getLength(); ++i) {
            if (branchAttributes.item(i).getNodeName().equals("name")) {
                branchType.setName(branchAttributes.item(i).getNodeValue());
            } else if (branchAttributes.item(i).getNodeName().equals("idltype")) {
                String idlType = branchAttributes.item(i).getNodeValue();
                int seperatorIndex = idlType.indexOf(':');
                String prefix = idlType.substring(0, seperatorIndex);
                String localPart = idlType.substring(seperatorIndex + 1, idlType.length());
                branchType.setIdltype(new QName(def.getNamespace(prefix), localPart, prefix));
            } else if (branchAttributes.item(i).getNodeName().equals("default")) {
                Boolean defaultValue = new Boolean(branchAttributes.item(i).getNodeValue());
                branchType.setDefault(defaultValue);
            }
        }
        
        NodeList branchChildNodes = node.getChildNodes();
        for (int i = 0; i < branchChildNodes.getLength(); ++i) {
            Node currentNode = branchChildNodes.item(i);
            
            if (currentNode.getNodeName().equals("corba:case")) {
                CaseType caseType = new CaseType();
                NamedNodeMap caseAttributes = currentNode.getAttributes();
                
                for (int j = 0; j < caseAttributes.getLength(); ++j) {
                    if (caseAttributes.item(j).getNodeName().equals("label")) {
                        caseType.setLabel(caseAttributes.item(j).getNodeValue());
                    }
                }
                branchType.getCase().add(caseType);
            }
        }
        
        return branchType;
    }
    
    public Alias getAliasDefinition(Node node, Definition def) {
        Alias aliasType = new Alias();

        // Store information about the typedef
        NamedNodeMap aliasAttributes = node.getAttributes();
        for (int i = 0; i < aliasAttributes.getLength(); ++i) {
            if (aliasAttributes.item(i).getNodeName().equals("name")) {
                aliasType.setName(aliasAttributes.item(i).getNodeValue());
            } else if (aliasAttributes.item(i).getNodeName().equals("repositoryID")) {
                aliasType.setRepositoryID(aliasAttributes.item(i).getNodeValue());
            } else if (aliasAttributes.item(i).getNodeName().equals("basetype")) {
                String baseType = aliasAttributes.item(i).getNodeValue();
                int seperatorIndex = baseType.indexOf(':');
                String prefix = baseType.substring(0, seperatorIndex);
                String localPart = baseType.substring(seperatorIndex + 1, baseType.length());
                aliasType.setBasetype(new QName(def.getNamespace(prefix), localPart, prefix));
            }
        }
        return aliasType;
    }

    public Array getArrayDefinition(Node node, Definition def) {
        Array arrayType = new Array();

        // Store information about the array
        NamedNodeMap arrayAttributes = node.getAttributes();
        for (int i = 0; i < arrayAttributes.getLength(); ++i) {
            if (arrayAttributes.item(i).getNodeName().equals("name")) {
                arrayType.setName(arrayAttributes.item(i).getNodeValue());
            } else if (arrayAttributes.item(i).getNodeName().equals("repositoryID")) {
                arrayType.setRepositoryID(arrayAttributes.item(i).getNodeValue());
            } else if (arrayAttributes.item(i).getNodeName().equals("elemtype")) {
                String elemType = arrayAttributes.item(i).getNodeValue();
                int seperatorIndex = elemType.indexOf(':');
                String prefix = elemType.substring(0, seperatorIndex);
                String localPart = elemType.substring(seperatorIndex + 1, elemType.length());
                arrayType.setElemtype(new QName(def.getNamespace(prefix), localPart, prefix));
            } else if (arrayAttributes.item(i).getNodeName().equals("bound")) {
                arrayType.setBound(Long.parseLong(arrayAttributes.item(i).getNodeValue()));
            }
        }
        return arrayType;
    }

    public Sequence getSequenceDefinition(Node node, Definition def) {
        Sequence sequenceType = new Sequence();

        // Store information about the sequence
        NamedNodeMap sequenceAttributes = node.getAttributes();
        for (int i = 0; i < sequenceAttributes.getLength(); ++i) {
            if (sequenceAttributes.item(i).getNodeName().equals("name")) {
                sequenceType.setName(sequenceAttributes.item(i).getNodeValue());
            } else if (sequenceAttributes.item(i).getNodeName().equals("repositoryID")) {
                sequenceType.setRepositoryID(sequenceAttributes.item(i).getNodeValue());
            } else if (sequenceAttributes.item(i).getNodeName().equals("elemtype")) {
                String elemType = sequenceAttributes.item(i).getNodeValue();
                int seperatorIndex = elemType.indexOf(':');
                String prefix = elemType.substring(0, seperatorIndex);
                String localPart = elemType.substring(seperatorIndex + 1, elemType.length());
                sequenceType.setElemtype(new QName(def.getNamespace(prefix), localPart, prefix));
            } else if (sequenceAttributes.item(i).getNodeName().equals("bound")) {
                sequenceType.setBound(Long.parseLong(sequenceAttributes.item(i).getNodeValue()));
            }
        }
        return sequenceType;
    }

    public Anonstring getAnonStringDefinition(Node node, Definition def) {
        Anonstring anonStringType = new Anonstring();

        // Store information about the fixed type
        NamedNodeMap anonStringAttributes = node.getAttributes();
        for (int i = 0; i < anonStringAttributes.getLength(); ++i) {
            if (anonStringAttributes.item(i).getNodeName().equals("name")) {
                anonStringType.setName(anonStringAttributes.item(i).getNodeValue());
            } else if (anonStringAttributes.item(i).getNodeName().equals("bound")) {
                anonStringType.setBound(new Long(anonStringAttributes.item(i).getNodeValue()));
            }
        }
        return anonStringType;
    }

    public Anonstring getAnonWStringDefinition(Node node, Definition def) {
        Anonstring anonStringType = new Anonstring();

        // Store information about the fixed type
        NamedNodeMap anonStringAttributes = node.getAttributes();
        for (int i = 0; i < anonStringAttributes.getLength(); ++i) {
            if (anonStringAttributes.item(i).getNodeName().equals("name")) {
                anonStringType.setName(anonStringAttributes.item(i).getNodeValue());
            } else if (anonStringAttributes.item(i).getNodeName().equals("bound")) {
                anonStringType.setBound(new Long(anonStringAttributes.item(i).getNodeValue()));
            }
        }
        return anonStringType;
    }

    public Anonfixed getAnonFixedDefinition(Node node, Definition def) {
        Anonfixed anonFixedType = new Anonfixed();

        // Store information about the fixed type
        NamedNodeMap anonFixedAttributes = node.getAttributes();
        for (int i = 0; i < anonFixedAttributes.getLength(); ++i) {
            if (anonFixedAttributes.item(i).getNodeName().equals("name")) {
                anonFixedType.setName(anonFixedAttributes.item(i).getNodeValue());
            } else if (anonFixedAttributes.item(i).getNodeName().equals("digits")) {
                anonFixedType.setDigits(new Long(anonFixedAttributes.item(i).getNodeValue()));
            } else if (anonFixedAttributes.item(i).getNodeName().equals("scale")) {
                anonFixedType.setScale(new Long(anonFixedAttributes.item(i).getNodeValue()));
            }
        }
        return anonFixedType;
    }

    public Anonsequence getAnonSequenceDefinition(Node node, Definition def) {
        Anonsequence anonSequenceType = new Anonsequence();

        // Store information about the sequence
        NamedNodeMap anonSequenceAttributes = node.getAttributes();
        for (int i = 0; i < anonSequenceAttributes.getLength(); ++i) {
            if (anonSequenceAttributes.item(i).getNodeName().equals("name")) {
                anonSequenceType.setName(anonSequenceAttributes.item(i).getNodeValue());
            } else if (anonSequenceAttributes.item(i).getNodeName().equals("elemtype")) {
                String elemType = anonSequenceAttributes.item(i).getNodeValue();
                int seperatorIndex = elemType.indexOf(':');
                String prefix = elemType.substring(0, seperatorIndex);
                String localPart = elemType.substring(seperatorIndex + 1, elemType.length());
                anonSequenceType.setElemtype(new QName(def.getNamespace(prefix), localPart, prefix));
            } else if (anonSequenceAttributes.item(i).getNodeName().equals("bound")) {
                anonSequenceType.setBound(Long.parseLong(anonSequenceAttributes.item(i).getNodeValue()));
            }
        }
        return anonSequenceType;
    }

    public Anonarray getAnonArrayDefinition(Node node, Definition def) {
        Anonarray anonArrayType = new Anonarray();

        // Store information about the array
        NamedNodeMap anonArrayAttributes = node.getAttributes();
        for (int i = 0; i < anonArrayAttributes.getLength(); ++i) {
            if (anonArrayAttributes.item(i).getNodeName().equals("name")) {
                anonArrayType.setName(anonArrayAttributes.item(i).getNodeValue());
            } else if (anonArrayAttributes.item(i).getNodeName().equals("elemtype")) {
                String elemType = anonArrayAttributes.item(i).getNodeValue();
                int seperatorIndex = elemType.indexOf(':');
                String prefix = elemType.substring(0, seperatorIndex);
                String localPart = elemType.substring(seperatorIndex + 1, elemType.length());
                anonArrayType.setElemtype(new QName(def.getNamespace(prefix), localPart, prefix));
            } else if (anonArrayAttributes.item(i).getNodeName().equals("bound")) {
                anonArrayType.setBound(Long.parseLong(anonArrayAttributes.item(i).getNodeValue()));
            }
        }
        return anonArrayType;
    }
}                
