/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba.types;

import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.schemas.yoko.bindings.corba.Alias;
import org.apache.schemas.yoko.bindings.corba.Anonarray;
import org.apache.schemas.yoko.bindings.corba.Anonsequence;
import org.apache.schemas.yoko.bindings.corba.Array;
import org.apache.schemas.yoko.bindings.corba.Exception;
import org.apache.schemas.yoko.bindings.corba.MemberType;
import org.apache.schemas.yoko.bindings.corba.Sequence;
import org.apache.schemas.yoko.bindings.corba.Struct;
import org.apache.schemas.yoko.bindings.corba.Union;
import org.apache.schemas.yoko.bindings.corba.Unionbranch;

import org.apache.yoko.bindings.corba.CorbaTypeMap;
import org.apache.yoko.bindings.corba.CorbaUtils;
import org.apache.yoko.wsdl.CorbaTypeImpl;

import org.omg.CORBA.ORB;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;

public class CorbaHandlerUtils {

    public static CorbaObjectHandler createTypeHandler(ORB orb, QName name, 
                                                       QName idlType, List<CorbaTypeMap> typeMaps) {
        CorbaObjectHandler handler = null;

        TypeCode tc = CorbaUtils.getTypeCode(orb, idlType, typeMaps);
        if (CorbaUtils.isPrimitiveIdlType(idlType)) {
            // Any is a special primitive
            if (tc.kind().value() == TCKind._tk_any) {
                handler = new CorbaAnyHandler(name, idlType, tc, null);                
            } else {
                handler = new CorbaPrimitiveHandler(name, idlType, tc, null);
            }
        } else {
            CorbaTypeImpl type = CorbaUtils.getCorbaType(idlType, typeMaps);

            switch (tc.kind().value()) {
            case TCKind._tk_alias:
                // For typedefs, we'll create the handler for the Object that is typedef'd.  Reading and
                // writing will use an Alias object and typecode, find the correct base type and then use
                // this handler to marshal/unmarshal data
                Alias aliasType = (Alias)type;
                handler = createTypeHandler(orb, name, aliasType.getBasetype(), typeMaps);
                break;
            case TCKind._tk_array:
                handler = new CorbaArrayHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_enum:
                handler = new CorbaEnumHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_except:
                handler = new CorbaExceptionHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_fixed:
                handler = new CorbaFixedHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_sequence:
                handler = new CorbaSequenceHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_struct:
                handler = new CorbaStructHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_union:
                handler = new CorbaUnionHandler(name, idlType, tc, type);
                break;
            case TCKind._tk_string:
            case TCKind._tk_wstring:
                // These need to be here to catch the anonymous string types.
                handler = new CorbaPrimitiveHandler(name, idlType, tc, type);
                break;
            default:
                handler = new CorbaObjectHandler(name, idlType, tc, type);                
            }
        }

        return handler;
    }
    
    public static CorbaObjectHandler initializeObjectHandler(ORB orb, QName name, 
                                                             QName idlType, List<CorbaTypeMap> typeMaps) {
        CorbaObjectHandler obj = createTypeHandler(orb, name, idlType, typeMaps);

        if (!CorbaUtils.isPrimitiveIdlType(idlType)) {
            switch (obj.getTypeCode().kind().value()) {
            case TCKind._tk_array:
                initializeArrayHandler(orb, obj, name, typeMaps);
                break;
            case TCKind._tk_except:
                initializeExceptionHandler(orb, obj, name, typeMaps);
                break;
            case TCKind._tk_sequence:
                initializeSequenceHandler(orb, obj, name, typeMaps);
                break;
            case TCKind._tk_struct:
                initializeStructHandler(orb, obj, name, typeMaps);
                break;
            case TCKind._tk_union:
                initializeUnionHandler(orb, obj, name, typeMaps);
                break;

            default:
               // TODO: Should we raise an exception or log?
            }
        }
        return obj;
    }
    
    public static void initializeArrayHandler(ORB orb, CorbaObjectHandler obj, 
                                              QName name, List<CorbaTypeMap> typeMaps) {
        QName arrayElementType = null;
        long arrayBound = 0;
        CorbaTypeImpl baseType = obj.getType();
        if (baseType instanceof Array) {
            Array arrayType = (Array)baseType;
            arrayElementType = arrayType.getElemtype();
            arrayBound = arrayType.getBound();
        } else {
            Anonarray anonArrayType = (Anonarray)baseType;
            arrayElementType = anonArrayType.getElemtype();
            arrayBound = anonArrayType.getBound();
        }
        for (int i = 0; i < arrayBound; ++i) {
            QName elementName = new QName(name.getNamespaceURI(), "item");
            CorbaObjectHandler elementObj = 
                initializeObjectHandler(orb, elementName, arrayElementType, typeMaps);
            ((CorbaArrayHandler)obj).addElement(elementObj);
        }
    }
    
    public static void initializeExceptionHandler(ORB orb, CorbaObjectHandler obj, 
                                                  QName name, List<CorbaTypeMap> typeMaps) {
        Exception exceptType = (Exception)obj.getType();
        List<MemberType> exceptMembers = exceptType.getMember();

        for (int i = 0; i < exceptMembers.size(); ++i) {
            MemberType member = exceptMembers.get(i);
            QName memberName = new QName(name.getNamespaceURI(), member.getName());
            QName memberType = member.getIdltype();
            CorbaObjectHandler memberObj = 
                initializeObjectHandler(orb, memberName, memberType, typeMaps);
            ((CorbaExceptionHandler)obj).addMember(memberObj);
        }
    }
    
    public static void initializeSequenceHandler(ORB orb, CorbaObjectHandler obj, 
                                                 QName name, List<CorbaTypeMap> typeMaps) {
        QName seqElementType = null;
        long seqBound = 0;
        CorbaTypeImpl baseType = obj.getType();
        if (baseType instanceof Sequence) {
            Sequence seqType = (Sequence)baseType;
            seqElementType = seqType.getElemtype();
            seqBound = seqType.getBound();
        } else {
            Anonsequence seqType = (Anonsequence)baseType;
            seqElementType = seqType.getElemtype();
            seqBound = seqType.getBound();
        }
        if (seqBound == 0) {
            // This is an unbounded sequence.  Store a 'template' object that we can use to create
            // new objects as needed
            QName elementName = new QName(name.getNamespaceURI(), "item");
            CorbaObjectHandler elementObj = 
                initializeObjectHandler(orb, elementName, seqElementType, typeMaps);
            ((CorbaSequenceHandler)obj).setTemplateElement(elementObj);
        }
        for (int i = 0; i < seqBound; ++i) {
            QName elementName = new QName(name.getNamespaceURI(), "item");
            CorbaObjectHandler elementObj = 
                initializeObjectHandler(orb, elementName, seqElementType, typeMaps);
            ((CorbaSequenceHandler)obj).addElement(elementObj);
        }
    }
    
    public static void initializeStructHandler(ORB orb, CorbaObjectHandler obj, 
                                               QName name, List<CorbaTypeMap> typeMaps) {
        Struct structType = (Struct)obj.getType();
        List<MemberType> structMembers = structType.getMember();

        for (int i = 0; i < structMembers.size(); ++i) {
            MemberType member = structMembers.get(i);
            QName memberName = new QName(name.getNamespaceURI(), member.getName());
            QName memberType = member.getIdltype();
            CorbaObjectHandler memberObj = 
                initializeObjectHandler(orb, memberName, memberType, typeMaps);
            ((CorbaStructHandler)obj).addMember(memberObj);
        }
    }
    
    public static void initializeUnionHandler(ORB orb, CorbaObjectHandler obj, 
                                              QName name, List<CorbaTypeMap> typeMaps) {
        Union unionType = (Union)obj.getType();
        // First handle the discriminator
        CorbaObjectHandler discObj = initializeObjectHandler(orb, 
                                                             new QName("discriminator"),
                                                             unionType.getDiscriminator(),
                                                             typeMaps);
        ((CorbaUnionHandler)obj).setDiscriminator(discObj);
        
        // Now handle all of the branches
        List<Unionbranch> unionBranches = unionType.getUnionbranch();
        for (Iterator<Unionbranch> iter = unionBranches.iterator(); iter.hasNext();) {
            Unionbranch branch = iter.next();
            QName branchName = new QName(name.getNamespaceURI(), branch.getName());
            QName branchIdlType = branch.getIdltype();
            CorbaObjectHandler branchObj = 
                initializeObjectHandler(orb, branchName, branchIdlType, typeMaps);
            ((CorbaUnionHandler)obj).addCase(branchObj);
        }
    }
}
