/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import javax.xml.namespace.QName;
import javax.xml.ws.Holder;

import org.apache.yoko.bindings.corba.types.CorbaPrimitiveHandler;
import org.apache.yoko.wsdl.CorbaConstants;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.BindingManager;
import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bus.jaxws.JAXBDataBindingCallback;
import org.objectweb.celtix.context.GenericMessageContext;
import org.objectweb.celtix.context.ObjectMessageContext;
import org.objectweb.celtix.context.ObjectMessageContextImpl;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;

import junit.framework.TestCase;

import org.omg.CORBA.ORB;
import org.omg.CORBA.TCKind;

public class CorbaBindingImplUnmarshalTest extends TestCase {

    private static String TEST_CLASS = "org.apache.schemas.yoko.idl.parammodes.ParamModes";
    
    private TestUtils testUtils;
    private ORB orb;
    private Bus bus;
    private EndpointReferenceType endpoint;
    private CorbaBindingImpl bindingImpl;
    private Class paramModesClass;
    
    public CorbaBindingImplUnmarshalTest(String arg0) {
        super(arg0);
    }
    public static void main(String[] args) {
        junit.textui.TestRunner.run(CorbaBindingImplUnmarshalTest.class);
    }
    
    protected void setUp() throws Exception {
        super.setUp();
        
        testUtils = new TestUtils();
    
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "org.apache.yoko.orb.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass", "org.apache.yoko.orb.CORBA.ORBSingleton");
        props.put("yoko.orb.id", "Yoko-Test");
        orb = ORB.init(new String[0], props);
        bus = Bus.init(new String[0]);
        
        paramModesClass = Thread.currentThread().getContextClassLoader().loadClass(TEST_CLASS);
        
        // programmatically add the corba binding factory
        BindingManager manager = bus.getBindingManager();
        CorbaBindingFactory corbaBF = new CorbaBindingFactory();
        corbaBF.init(bus);
        manager.registerBinding(CorbaConstants.NU_WSDL_CORBA, corbaBF);
        
        endpoint = testUtils.getParamModesEndpointReference();
        bindingImpl = new CorbaBindingImpl(bus, endpoint, orb, true);
    }
    
    protected void tearDown() throws Exception {
        if (orb != null) {
            try {
                orb.destroy();
            } catch (Exception ex) {
                // Do nothing.  Throw an Exception?
            }
        }
    }
    
    public void testUnmarshalSingleInParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[1];
        params[0] = short.class;
        
        CorbaPrimitiveHandler handler = new CorbaPrimitiveHandler(new QName("p"),
                                                                  CorbaConstants.NT_CORBA_SHORT,
                                                                  orb.get_primitive_tc(TCKind.tk_short),
                                                                  null);
        handler.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler, handler.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInParam", params));
        objContext.setMessageObjects((short)0);
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 1);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInParamWithReturn", params));
        objContext.setMessageObjects((short)0);
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 1);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
   }

    public void testUnmarshalSingleOutParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[0];
        
        CorbaPrimitiveHandler handler = new CorbaPrimitiveHandler(new QName("p"),
                                                                  CorbaConstants.NT_CORBA_SHORT,
                                                                  orb.get_primitive_tc(TCKind.tk_short),
                                                                  null);
        handler.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler, handler.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleOutParam", params));
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        // In this case, there should be no parameters since the out param is mapped to the return type
        assertTrue(resultArgs == null);
        
        // Now try the call with a return type
        params = new Class[1];
        params[0] = new Holder<Short>().getClass();
        
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleOutParamWithReturn", params));
        objContext.setMessageObjects(new Holder<Short>(new Short((short)0)));
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 1);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
   }

    public void testUnmarshalSingleInOutParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[1];
        params[0] = new Holder<Short>().getClass();
        
        CorbaPrimitiveHandler handler = new CorbaPrimitiveHandler(new QName("p"),
                                                                  CorbaConstants.NT_CORBA_SHORT,
                                                                  orb.get_primitive_tc(TCKind.tk_short),
                                                                  null);
        handler.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler, handler.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInOutParam", params));
        objContext.setMessageObjects(new Holder<Short>(new Short((short)0)));
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 1);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);

        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInOutParamWithReturn", params));
        objContext.setMessageObjects(new Holder<Short>(new Short((short)0)));
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 1);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
   }

    public void testUnmarshalMultipleInParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[2];
        params[0] = short.class;
        params[1] = short.class;
        
        CorbaPrimitiveHandler handler0 = new CorbaPrimitiveHandler(new QName("p1"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler0.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler0, handler0.getName());
        CorbaPrimitiveHandler handler1 = new CorbaPrimitiveHandler(new QName("p2"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler1.setValue(new Short((short)456));
        CorbaStreamable arg1 = new CorbaStreamable(handler1, handler1.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        corbaContext.getMessage().addStreamableArgument(arg1);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInParams", params));
        Object[] objCtxArgs = new Object[2];
        objCtxArgs[0] = (short)0;
        objCtxArgs[1] = (short)0;
        objContext.setMessageObjects(objCtxArgs);
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);

        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInParamsWithReturn", params));
        objContext.setMessageObjects(objCtxArgs);
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
   }

    public void testUnmarshalMultipleOutParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[2];
        params[0] = new Holder<Short>().getClass();
        params[1] = new Holder<Short>().getClass();
        
        CorbaPrimitiveHandler handler0 = new CorbaPrimitiveHandler(new QName("p1"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler0.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler0, handler0.getName());
        CorbaPrimitiveHandler handler1 = new CorbaPrimitiveHandler(new QName("p2"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler1.setValue(new Short((short)456));
        CorbaStreamable arg1 = new CorbaStreamable(handler1, handler1.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        corbaContext.getMessage().addStreamableArgument(arg1);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleOutParams", params));
        Object[] objCtxArgs = new Object[2];
        objCtxArgs[0] = new Holder<Short>(new Short((short)0));
        objCtxArgs[1] = new Holder<Short>(new Short((short)0));
        objContext.setMessageObjects(objCtxArgs);
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleOutParamsWithReturn", params));
        objContext.setMessageObjects(objCtxArgs);
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
    }

    public void testUnmarshalMultipleInOutParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[2];
        params[0] = new Holder<Short>().getClass();
        params[1] = new Holder<Short>().getClass();
        
        CorbaPrimitiveHandler handler0 = new CorbaPrimitiveHandler(new QName("p1"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler0.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler0, handler0.getName());
        CorbaPrimitiveHandler handler1 = new CorbaPrimitiveHandler(new QName("p2"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler1.setValue(new Short((short)456));
        CorbaStreamable arg1 = new CorbaStreamable(handler1, handler1.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        corbaContext.getMessage().addStreamableArgument(arg1);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInOutParams", params));
        Object[] objCtxArgs = new Object[2];
        objCtxArgs[0] = new Holder<Short>(new Short((short)0));
        objCtxArgs[1] = new Holder<Short>(new Short((short)0));
        objContext.setMessageObjects(objCtxArgs);
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);

        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInOutParamsWithReturn", params));
        objContext.setMessageObjects(objCtxArgs);
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 2);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
    }

    public void testUnmarshalMultipleMixedParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // Prepare the arguments
        params = new Class[3];
        params[0] = short.class;
        params[1] = new Holder<Short>().getClass();
        params[2] = new Holder<Short>().getClass();
        
        
        CorbaPrimitiveHandler handler0 = new CorbaPrimitiveHandler(new QName("p1"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler0.setValue(new Short((short)123));
        CorbaStreamable arg0 = new CorbaStreamable(handler0, handler0.getName());
        CorbaPrimitiveHandler handler1 = new CorbaPrimitiveHandler(new QName("p2"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler1.setValue(new Short((short)456));
        CorbaStreamable arg1 = new CorbaStreamable(handler1, handler1.getName());
        CorbaPrimitiveHandler handler2 = new CorbaPrimitiveHandler(new QName("p3"),
                                                                   CorbaConstants.NT_CORBA_SHORT,
                                                                   orb.get_primitive_tc(TCKind.tk_short),
                                                                   null);
        handler2.setValue(new Short((short)789));
        CorbaStreamable arg2 = new CorbaStreamable(handler2, handler2.getName());

        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        corbaContext.setMessage(new CorbaMessage());
        corbaContext.getMessage().addStreamableArgument(arg0);
        corbaContext.getMessage().addStreamableArgument(arg1);
        corbaContext.getMessage().addStreamableArgument(arg2);
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleMixedParams", params));
        Object[] objCtxArgs = new Object[3];
        objCtxArgs[0] = (short)0;
        objCtxArgs[1] = new Holder<Short>(new Short((short)0));
        objCtxArgs[2] = new Holder<Short>(new Short((short)0));
        objContext.setMessageObjects(objCtxArgs);
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        Object[] resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 3);
        
        // Now try the call with a return type
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, false);
        corbaContext.getMessage().setStreamableReturn(arg0);

        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleMixedParamsWithReturn", params));
        objContext.setMessageObjects(objCtxArgs);
        
        bindingImpl.unmarshal(corbaContext,
                              objContext,
                              new JAXBDataBindingCallback(objContext.getMethod(),
                                                          DataBindingCallback.Mode.MESSAGE,
                                                          null));
        
        resultArgs = objContext.getMessageObjects();
        assertTrue(resultArgs.length == 3);
        Object ret = objContext.getReturn();
        assertNotNull(ret);
    }
}
