/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.wsdl.BindingFault;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Definition;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.Part;

import javax.xml.namespace.QName;

import antlr.ASTVisitor;
import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.RaisesType;

import org.apache.ws.commons.schema.XmlSchemaSequence;

public class MessageVisitor implements ASTVisitor {

    static final String IN_PARAMETER = "inparameter";
    static final String OUT_PARAMETER = "outparameter";
    static final String INOUT_PARAMETER = "inoutparameter";

    Definition definition;

    Operation operation;
    BindingOperation bindingOperation;
    OperationType corbaOperation;
    MessagePartVisitor partVisitor;
    ReturnParameterVisitor returnParameterVisitor;
    TypesVisitor typesVisitor;
    
    Message inputMessage;
    Message outputMessage;
    
    XmlSchemaSequence inWrappingSequence;
    XmlSchemaSequence outWrappingSequence;
   
    
    public MessageVisitor(WSDLASTVisitor visitor) {
        definition = visitor.getDefinition();
        outWrappingSequence = new XmlSchemaSequence();
        inWrappingSequence = new XmlSchemaSequence();
        partVisitor = new MessagePartVisitor(visitor, inWrappingSequence, outWrappingSequence);
        returnParameterVisitor = new ReturnParameterVisitor(visitor, outWrappingSequence);
        typesVisitor = visitor.getTypesVisitor();
    }

    public void setOperation(Operation op) {
        operation = op;
    }

    public void setBindingOperation(BindingOperation op) {
        bindingOperation = op;
        corbaOperation = (OperationType) op.getExtensibilityElements().get(0);
    }

    private AST getNextMessagePartNode(AST node) {
        AST currentNode = node;
        if (currentNode.getType() == IDLTokenTypes.LITERAL_unsigned) {
            currentNode = currentNode.getNextSibling();
        }
        if (currentNode.getType() == IDLTokenTypes.LITERAL_long
            && currentNode.getNextSibling().getType() == IDLTokenTypes.LITERAL_long) {
            currentNode = currentNode.getNextSibling();
        }
        return currentNode.getNextSibling();
    }
    
    public void visit(AST node) {
        AST node2 = node.getFirstChild();
        while (node2 != null) {
            switch (node2.getType()) {
            case IDLTokenTypes.LITERAL_in: {
                if (partVisitor.getInputMessage() == null) {                
                    partVisitor.setInputMessage(inputMessage);
                    partVisitor.setCorbaOperation(corbaOperation);
                }
                partVisitor.visit(node2);
                break;
            }
            case IDLTokenTypes.LITERAL_inout: {
                if (partVisitor.getInputMessage() == null) {                
                    partVisitor.setInputMessage(inputMessage);
                    partVisitor.setCorbaOperation(corbaOperation);
                }
                if (partVisitor.getOutputMessage() == null) {               
                    partVisitor.setOutputMessage(outputMessage);
                    partVisitor.setCorbaOperation(corbaOperation);
                }
                partVisitor.visit(node2);
                break;
            }
            case IDLTokenTypes.LITERAL_out: {
                if (partVisitor.getOutputMessage() == null) {               
                    partVisitor.setOutputMessage(outputMessage);
                    partVisitor.setCorbaOperation(corbaOperation);
                }
                partVisitor.visit(node2);
                break;
            }
            case IDLTokenTypes.LITERAL_void: {
                // nothing to do here
                break;
            }
            case IDLTokenTypes.LITERAL_raises: {
                //create a fault message
                createFaultMessage(node2.getFirstChild());
                break;
            }
            case IDLTokenTypes.LITERAL_oneway: {
                // TBD: oneway idl operations map to 
                // wsdl operations with no output message
                break;
            }
            default: {
                if (returnParameterVisitor.getOutputMessage() == null) {               
                    returnParameterVisitor.setOutputMessage(outputMessage);
                    returnParameterVisitor.setCorbaOperation(corbaOperation);
                }
                returnParameterVisitor.visit(node2);
            }
            }
            node2 = getNextMessagePartNode(node2);
        }
        boolean emptyMessage = false;
        if (partVisitor.getInputMessage() == null) {
            partVisitor.setInputMessage(inputMessage);
            emptyMessage = true;
        }
        if (partVisitor.getOutputMessage() == null) {               
            partVisitor.setOutputMessage(outputMessage);
            emptyMessage = true;
        }
        if (returnParameterVisitor.getOutputMessage() == null) {
            emptyMessage = true;
        }
        if (emptyMessage) {
            //try to visit the part visitor & create a empty sequence complex type.
        }

    }

    public void createInputMessage() {
        // message
        if (inputMessage == null) {
            Input input = definition.createInput();
            Message msg = definition.createMessage();
            msg.setQName(new QName(definition.getTargetNamespace(), operation.getName()));
            input.setMessage(msg);
            String msgName = operation.getName() + "Request";
            input.setName(msgName);
            msg.setUndefined(false);
            operation.setInput(input);
            definition.addMessage(msg);
            
            BindingInput bindingInput = definition.createBindingInput();
            bindingInput.setName(msgName);        
            bindingOperation.setBindingInput(bindingInput);
            
            inputMessage = msg;
        }
        
        // message - part
        Part part;
        QName element = inputMessage.getQName();
        if (inputMessage.getParts().size() == 0) {
            part = definition.createPart();
            part.setName(IN_PARAMETER);
            part.setElementName(element);
            typesVisitor.addWrapper(element, inWrappingSequence);
            inputMessage.addPart(part);
        } else {
            part = (Part) inputMessage.getParts().get(0);
        }        

    }

    public void createOutputMessage() {
        // message
        if  (outputMessage == null) {
            Output output = definition.createOutput();
            Message msg = definition.createMessage();
            msg.setQName(new QName(definition.getTargetNamespace(), operation.getName() + "Response"));
            output.setMessage(msg);
            String msgName = operation.getName() + "Response";
            output.setName(msgName);
            msg.setUndefined(false);
            operation.setOutput(output);
            definition.addMessage(msg);
    
            BindingOutput bindingOutput = definition.createBindingOutput();
            bindingOutput.setName(msgName);
            bindingOperation.setBindingOutput(bindingOutput);
            
            outputMessage = msg;
        }
        
        // message - part
        Part part;
        QName element = outputMessage.getQName();
        if (outputMessage.getParts().size() == 0) {
            part = definition.createPart();
            part.setName(OUT_PARAMETER);
            part.setElementName(element);
            typesVisitor.addWrapper(element, outWrappingSequence);
            outputMessage.addPart(part);
        } else {
            part = (Part) outputMessage.getParts().get(0);
        }

    }
    
    private void createFaultMessage(AST node) {
        while (node != null) {
            String exceptionName = node.toString();

            // message
            Message faultMsg = definition.createMessage();
            faultMsg.setQName(new QName(definition.getTargetNamespace(), exceptionName));
            faultMsg.setUndefined(false);

            // message - part
            Part part = definition.createPart();
            part.setName("exception");
            // REVISIT: should be reading QName from exception XmlSchemaElement
            // part.setElementName(element.getQName());
            part.setElementName(new QName(definition.getTargetNamespace(), exceptionName));
            faultMsg.addPart(part);

            // porttype - operation - fault
            Fault fault = definition.createFault();
            fault.setMessage(faultMsg);
            fault.setName(faultMsg.getQName().getLocalPart());
            operation.addFault(fault);

            // binding - operation - corba:operation - corba:raises
            RaisesType raisesType = new RaisesType();
            raisesType.setException(new QName(typesVisitor.getCorbaTypeMapping().getTargetNamespace(),
                                              exceptionName));
            corbaOperation.getRaises().add(raisesType);

            // binding - operation - fault
            BindingFault bindingFault = definition.createBindingFault();
            bindingFault.setName(faultMsg.getQName().getLocalPart());
            bindingOperation.addBindingFault(bindingFault);

            definition.addMessage(faultMsg);
            
            node = node.getNextSibling();
        }
    }
}
