/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import java.util.Iterator;

import javax.xml.namespace.QName;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.TypeMappingType;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.apache.yoko.tools.common.CorbaPrimitiveMap;
import org.apache.yoko.tools.common.XmlSchemaPrimitiveMap;
import org.apache.yoko.wsdl.CorbaConstants;
import org.apache.yoko.wsdl.CorbaTypeImpl;

public final class TypesUtils {

    static CorbaPrimitiveMap corbaPrimitiveMap = new CorbaPrimitiveMap();
    static XmlSchemaPrimitiveMap xmlSchemaPrimitiveMap = new XmlSchemaPrimitiveMap();

    private TypesUtils() {
        //complete
    }

    public static XmlSchemaType findType(XmlSchemaCollection schemas, XmlSchema schema, AST node) {
        XmlSchemaType result = null; 
        QName corbaType = getPrimitiveType(node);
        if (corbaType != null) {
            QName schemaType = xmlSchemaPrimitiveMap.get(corbaType);
            result = schemas.getTypeByQName(schemaType);
        }
        if (result == null) {
            //REVISIT, properly reference the node name, need to look at the siblings.
            QName qname = new QName(schema.getTargetNamespace(), node.toString());
            result = schema.getTypeByName(qname);
        }
        return result;
    }

    public static CorbaTypeImpl findCorbaType(TypeMappingType typeMap, QName schemaTypeName) {
        CorbaTypeImpl result = (CorbaTypeImpl) corbaPrimitiveMap.get(schemaTypeName);
        if (result == null) {
            Iterator corbaTypes = typeMap.getStructOrExceptionOrUnion().iterator();
            while (corbaTypes.hasNext()) {
                CorbaTypeImpl type = (CorbaTypeImpl) corbaTypes.next();
                if (type.getQName().getLocalPart().equals(schemaTypeName.getLocalPart())) {
                    result = type;
                    break;
                }
            }
        }
        return result;
    }

    public static QName findCorbaTypeName(TypeMappingType typeMap, AST node) {
        return getPrimitiveType(node);
    }
    
    /** Returns node corresponding to the name of the CORBA primitive type node.
     * 
     * @param node
     * @return
     */
    public static AST getPrimitiveCorbaTypeNameNode(AST node) {
        AST currentNode = node;
        if (currentNode.getType() == IDLTokenTypes.LITERAL_unsigned) {
            currentNode = currentNode.getNextSibling();
        }
        if (currentNode.getType() == IDLTokenTypes.LITERAL_long
            && (currentNode.getNextSibling() != null)
            && (currentNode.getNextSibling().getType() == IDLTokenTypes.LITERAL_long)) {
            currentNode = currentNode.getNextSibling();
        }
        return currentNode.getNextSibling();
    }

    private static QName getPrimitiveType(AST node) {
        QName result = null;
        switch (node.getType()) {
        case IDLTokenTypes.LITERAL_long:
            if ((node.getNextSibling() != null)
                && (node.getNextSibling().getType() == IDLTokenTypes.LITERAL_long)) {
                // long long
                result = CorbaConstants.NT_CORBA_LONGLONG;
            } else if ((node.getFirstChild() != null)
                && (node.getFirstChild().getType() == IDLTokenTypes.LITERAL_double)) {
                // "double" node is a child of "long" node - instead of being a sibling
                // long double
                result = CorbaConstants.NT_CORBA_LONGDOUBLE;
            } else {
                // long
                result = CorbaConstants.NT_CORBA_LONG;
            }
            break;
        case IDLTokenTypes.LITERAL_unsigned:
            AST node2 = node.getNextSibling();
            if (node2 != null && node2.getType() == IDLTokenTypes.LITERAL_short) {
                // unsigned short
                result = CorbaConstants.NT_CORBA_USHORT;
            } else if (node2 != null && node2.getType() == IDLTokenTypes.LITERAL_long) {
                AST node3 = node2.getNextSibling();
                if (node3 != null && node3.getType() == IDLTokenTypes.LITERAL_long) {
                    // unsigned long long
                    result = CorbaConstants.NT_CORBA_ULONGLONG;
                } else {
                    // unsigned long
                    result = CorbaConstants.NT_CORBA_ULONG;
                }
            } else {
                // TBD: we should never get here
            }
            break;
        case IDLTokenTypes.LITERAL_short:
            result = CorbaConstants.NT_CORBA_SHORT;
            break;
        case IDLTokenTypes.LITERAL_float:
            result = CorbaConstants.NT_CORBA_FLOAT;
            break;            
        case IDLTokenTypes.LITERAL_double:
            result = CorbaConstants.NT_CORBA_DOUBLE;
            break;            
        case IDLTokenTypes.LITERAL_char:
            result = CorbaConstants.NT_CORBA_CHAR;
            break;
        case IDLTokenTypes.LITERAL_wchar:
            result = CorbaConstants.NT_CORBA_WCHAR;
            break;
        case IDLTokenTypes.LITERAL_string:
            result = CorbaConstants.NT_CORBA_STRING;
            break;
        case IDLTokenTypes.LITERAL_wstring:
            result = CorbaConstants.NT_CORBA_WSTRING;
            break;
        case IDLTokenTypes.LITERAL_boolean:
            result = CorbaConstants.NT_CORBA_BOOLEAN;
            break;
        case IDLTokenTypes.LITERAL_octet:
            result = CorbaConstants.NT_CORBA_OCTET;
            break;
        case IDLTokenTypes.LITERAL_any:
            result = CorbaConstants.NT_CORBA_ANY;
            break;
        default:
            // TBD 
            break;
        }
        return result;
    }

}
