/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import java.util.ArrayList;
import java.util.List;

import javax.wsdl.Definition;
import javax.wsdl.Types;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.schema.Schema;
import javax.wsdl.factory.WSDLFactory;

import javax.xml.namespace.QName;

import antlr.ASTVisitor;
import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.ArgType;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaObject;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaSerializer;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.apache.yoko.wsdl.CorbaConstants;
import org.apache.yoko.wsdl.CorbaTypeImpl;

public class TypesVisitor implements ASTVisitor {
    
    static final List<Integer> PRIMITIVE_TYPES = new ArrayList<Integer>();
    static final int PRIMITIVE = 0;

    TypeMappingType typeMap;

    Definition definition;
    WSDLFactory wsdlFactory;

    WSDLASTVisitor wsdlASTVisitor;
    
    XmlSchema schema;
    XmlSchemaCollection schemas;
    XmlSchemaObject currentType;

    ArgType currentParam;

    List<String> parts = new ArrayList<String>();

    public TypesVisitor(WSDLASTVisitor visitor)
        throws WSDLException {
        wsdlASTVisitor = visitor;
        definition = visitor.getDefinition();
        schemas = new XmlSchemaCollection();
        schema = new XmlSchema(definition.getTargetNamespace(), schemas);       
        createCorbaTypeMap();
    }

    public TypeMappingType getCorbaTypeMapping() {
        return typeMap;
    }
    
    public void attachSchema() throws Exception {
        Types types = definition.createTypes();
        Schema wsdlSchema = (Schema) 
            definition.getExtensionRegistry().createExtension(Types.class,
                                                              new QName("http://www.w3.org/2001/XMLSchema",
                                                                        "schema"));
        org.w3c.dom.Element el = XmlSchemaSerializer.serializeSchema(schema, true)[0].getDocumentElement();
        wsdlSchema.setElement(el);
        types.addExtensibilityElement(wsdlSchema);
        definition.setTypes(types);
    }
    
    /*-
     * Build the Wrapped Document Style wrapping elements
     * i.e. <xs:element name="...">
     *       <xs:complexType>
     *        <xs:sequence>
     *         ...
     *        </xs:sequence>
     *       </xs:complexType>
     *      </xs:element>
     */
    public void addWrapper(QName el, XmlSchemaSequence wrappingSequence) {
        XmlSchemaComplexType schemaComplexType = new XmlSchemaComplexType(schema);
        schemaComplexType.setParticle(wrappingSequence);
        
        XmlSchemaElement wrappingSchemaElement = new XmlSchemaElement();
        wrappingSchemaElement.setQName(el);
        wrappingSchemaElement.setName(el.getLocalPart());
        wrappingSchemaElement.setSchemaType(schemaComplexType);

        schema.getElements().add(el, wrappingSchemaElement);
        schema.getItems().add(wrappingSchemaElement);
    }

    public XmlSchemaElement addElement(XmlSchemaSequence schemaSequence, QName el, String name) {
        XmlSchemaElement element = new XmlSchemaElement();
        element.setQName(el);
        element.setName(name);

        schemaSequence.getItems().add(element);
        
        return element;
    }

    public void setCurrentPart(XmlSchemaElement element, String name, ArgType param) {
        //get the element & check if a type has been set
        if (element != null && element.getSchemaType() == null) {
            currentType = element;
            parts.add(name);
            currentParam = param;
        }
    }

    public void visit(AST node) {
        int type = getType(node);
        TypesVisitorBase visitor = null;
        switch (type) {
        case PRIMITIVE: {
            visitor = new PrimitiveTypesVisitor(schemas, schema, typeMap);
            visitor.visit(node);
            if (currentType instanceof XmlSchemaElement) {
                ((XmlSchemaElement) currentType).setSchemaTypeName(visitor.getSchemaTypeName());
                currentParam.setIdltype(visitor.getCorbaTypeName());
            }
            break;
        }
        case IDLTokenTypes.LITERAL_typedef: {
            visitor = new TypedefVisitor(schemas, schema, typeMap);
            visitor.visit(node);
            break;
        }
        case IDLTokenTypes.LITERAL_struct: {
            visitor = new StructVisitor(schemas, schema, typeMap);
            visitor.visit(node);
            break;
        }
        case IDLTokenTypes.LITERAL_exception: {
            visitor = new ExceptionVisitor(schemas, schema, typeMap);
            visitor.visit(node);
            break;
        }
        case IDLTokenTypes.LITERAL_const:
            ConstVisitor constVisitor = new ConstVisitor(schemas, schema, typeMap);
            constVisitor.visit(node);
            break;
        case IDLTokenTypes.LITERAL_enum:
            EnumVisitor enumVisitor = new EnumVisitor(schemas, schema, typeMap);
            enumVisitor.visit(node);
            break;
        case IDLTokenTypes.LITERAL_union:
            UnionVisitor unionVisitor = new UnionVisitor(schemas, schema, typeMap);
            unionVisitor.visit(node);
            break;
        case IDLTokenTypes.IDENT: {
            XmlSchemaType stype = TypesUtils.findType(schemas, schema, node);
            if (currentType instanceof XmlSchemaElement) {
                QName qname = stype.getQName();
                ((XmlSchemaElement) currentType).setSchemaTypeName(qname);
                CorbaTypeImpl corbaType = TypesUtils.findCorbaType(typeMap, qname);
                currentParam.setIdltype(corbaType.getQName());
            }
            break;
        }
        default: {
            System.err.println("Type not yet supported : " + node);
        }
        }

    }

    private int getType(AST node) {
        if (isPrimitive(node.getType())) {
            return PRIMITIVE;
        } else {
            return node.getType();
        }
    }

    private boolean isPrimitive(int nodeType) {
        return PRIMITIVE_TYPES.contains(nodeType);
    }

    private void createCorbaTypeMap() throws WSDLException { 
        typeMap = (TypeMappingType)
            definition.getExtensionRegistry().createExtension(Definition.class,
                                                              CorbaConstants.NE_CORBA_TYPEMAPPING);
        typeMap.setTargetNamespace(definition.getTargetNamespace()
                                  + "/"
                                  + CorbaConstants.NS_CORBA_TYPEMAP);
        definition.addExtensibilityElement(typeMap);
    }

    static {
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_float));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_double));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_long));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_short));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_unsigned));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_char));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_wchar));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_boolean));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_any));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_octet));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_string));
        PRIMITIVE_TYPES.add(new Integer(IDLTokenTypes.LITERAL_wstring));
    }

}
